"""This module contains utility functions and classes used in ``fluke``."""
from __future__ import annotations

import importlib
import inspect
import os
import sys
import warnings
from itertools import product
from typing import (TYPE_CHECKING, Any, Generator, Iterable, Literal, Optional,
                    Union)

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import psutil
import rich
import seaborn as sns
import torch
import yaml
from torch.nn import Module
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler

sys.path.append(".")
sys.path.append("..")

if TYPE_CHECKING:
    from client import Client  # NOQA
    from server import Server  # NOQA

from .. import DDict, FlukeCache, FlukeENV  # NOQA

__all__ = [
    'log',
    'model',
    'Configuration',
    'ClientObserver',
    'OptimizerConfigurator',
    'ServerObserver',
    'bytes2human',
    'cache_obj',
    'clear_cuda_cache',
    'get_class_from_str',
    'get_class_from_qualified_name',
    'get_full_classname',
    'get_loss',
    'get_model',
    'get_optimizer',
    'get_scheduler',
    'flatten_dict',
    'import_module_from_str',
    'memory_usage',
    'plot_distribution',
    'retrieve_obj'
]


class ClientObserver():
    """Client observer interface.
    This interface is used to observe the client during the federated learning process.
    For example, it can be used to log the performance of the local model, as it is done by the
    :class:`fluke.utils.log.Log` class.
    """

    def start_fit(self, round: int, client_id: int, model: Module, **kwargs: dict[str, Any]):
        """This method is called when the client starts the local training process.

        Args:
            round (int): The round number.
            client_id (int):  The client ID.
            model (Module): The local model before training.
            **kwargs (dict): Additional keyword arguments.
        """
        pass

    def end_fit(self,
                round: int,
                client_id: int,
                model: Module,
                loss: float,
                **kwargs: dict[str, Any]):
        """This method is called when the client ends the local training process.

        Args:
            round (int): The round number.
            client_id (int): The client ID.
            model (Module): The local model after the local training.
            loss (float): The average loss incurred by the local model during training.
            **kwargs (dict): Additional keyword arguments.
        """
        pass

    def client_evaluation(self,
                          round: int,
                          client_id: int,
                          phase: Literal["pre-fit", "post-fit"],
                          evals: dict[str, float],
                          **kwargs: dict[str, Any]):
        """This method is called when the client evaluates the local model.
        The evaluation can be done before ('pre-fit') and/or after ('post-fit') the local
        training process. The 'pre-fit' evlauation is usually the evaluation of the global model on
        the local test set, and the 'post-fit' evaluation is the evaluation of the just updated
        local model on the local test set.

        Args:
            round (int): The round number.
            client_id (int): The client ID.
            phase (Literal['pre-fit', 'post-fit']): Whether the evaluation is done before or after
                the local training process.
            evals (dict[str, float]): The evaluation results.
            **kwargs (dict): Additional keyword arguments.
        """
        pass


class ServerObserver():
    """Server observer interface.
    This interface is used to observe the server during the federated learning process.
    For example, it can be used to log the performance of the global model and the communication
    costs, as it is done by the ``Log`` class.
    """

    def start_round(self, round: int, global_model: Any) -> None:
        """This method is called when a new round starts.

        Args:
            round (int): The round number.
            global_model (Any): The current global model.
        """
        pass

    def end_round(self, round: int) -> None:
        """This method is called when a round ends.

        Args:
            round (int): The round number.
        """
        pass

    def selected_clients(self, round: int, clients: Iterable) -> None:
        """This method is called when the clients have been selected for the current round.

        Args:
            round (int): The round number.
            clients (Iterable): The clients selected for the current round.
        """
        pass

    def server_evaluation(self,
                          round: int,
                          type: Literal["global", "locals"],
                          evals: Union[dict[str, float], dict[int, dict[str, float]]],
                          **kwargs: dict[str, Any]) -> None:
        """This method is called when the server evaluates the global or the local models on its
        test set.

        Args:
            round (int): The round number.
            type (Literal['global', 'locals']): The type of evaluation. If 'global', the evaluation
                is done on the global model. If 'locals', the evaluation is done on the local models
                of the clients on the test set of the server.
            evals (dict[str, float] | dict[int, dict[str, float]]): The evaluation metrics. In case
                of 'global' evaluation, it is a dictionary with the evaluation metrics. In case of
                'locals' evaluation, it is a dictionary of dictionaries where the keys are the
                client IDs and the values are the evaluation metrics.
        """
        pass

    def finished(self, round: int) -> None:
        """This method is called when the federated learning process has ended.

        Args:
            round (int): The last round number.
        """
        pass

    def interrupted(self) -> None:
        """This method is called when the federated learning process has been interrupted."""
        pass


class OptimizerConfigurator:
    """This class is used to configure the optimizer and the learning rate scheduler.

    Attributes:
        optimizer (type[Optimizer]): The optimizer class.
        scheduler (type[LRScheduler]): The learning rate scheduler class.
        optimizer_cfg (DDict): The optimizer keyword arguments.
        scheduler_cfg (DDict): The scheduler keyword arguments.
    """

    def __init__(self,
                 optimizer_cfg: DDict | dict,
                 scheduler_cfg: DDict | dict = None):
        """Initialize the optimizer configurator. In both the ``optimizer_cfg`` and the
        ``scheduler_cfg`` dictionaries, the key "name" is used to specify the optimizer and the
        scheduler, respectively. If not present, the default optimizer is the ``SGD`` optimizer, and
        the default scheduler is the ``StepLR`` scheduler. The other keys are the keyword arguments
        for the optimizer and the scheduler, respectively.

        Note:
            In the key "name" of the ``optimizer_cfg`` dictionary, you can specify the optimizer
            class directly, instead of its string name. Same for the scheduler.

        Args:
            optimizer_cfg (dict, DDict): The optimizer class.
            scheduler_cfg (dict or DDict, optional): The scheduler keyword arguments. If not
                specified, the default scheduler is the ``StepLR`` scheduler with a step size of 1
                and a gamma of 1. Defaults to ``None``.
        """

        self.optimizer_cfg: DDict = None
        self.scheduler_cfg: DDict = None
        self.optimizer: type[Optimizer] = None
        self.scheduler: type[LRScheduler] = None

        if isinstance(optimizer_cfg, DDict):
            self.optimizer_cfg = optimizer_cfg
        elif isinstance(optimizer_cfg, dict):
            self.optimizer_cfg = DDict(**optimizer_cfg)
        else:
            raise ValueError("Invalid optimizer configuration.")

        if scheduler_cfg is None:
            self.scheduler_cfg = DDict(name="StepLR", step_size=1, gamma=1)
        elif isinstance(scheduler_cfg, DDict):
            self.scheduler_cfg = scheduler_cfg
        elif isinstance(scheduler_cfg, dict):
            self.scheduler_cfg = DDict(**scheduler_cfg)
        else:
            raise ValueError("Invalid scheduler configuration.")

        if isinstance(self.optimizer_cfg.name, str):
            self.optimizer = get_optimizer(self.optimizer_cfg.name)
        elif inspect.isclass(self.optimizer_cfg.name) and \
                issubclass(self.optimizer_cfg.name, Optimizer):
            self.optimizer = self.optimizer_cfg.name
        else:
            raise ValueError("Invalid optimizer name. Must be a string or an optimizer class.")

        if "name" not in self.scheduler_cfg:
            self.scheduler = get_scheduler("StepLR")
        elif isinstance(self.scheduler_cfg.name, str):
            self.scheduler = get_scheduler(self.scheduler_cfg.name)
        elif inspect.isclass(self.scheduler_cfg.name) and \
                issubclass(self.scheduler_cfg.name, LRScheduler):
            self.scheduler = self.scheduler_cfg.name
        else:
            raise ValueError("Invalid scheduler name. Must be a string or a scheduler class.")

        self.scheduler_cfg = self.scheduler_cfg.exclude("name")
        self.optimizer_cfg = self.optimizer_cfg.exclude("name")

    def __call__(self, model: Module, filter_fun: Optional[callable] = None, **override_kwargs):
        """Creates the optimizer and the scheduler.

        Args:
            model (Module): The model whose parameters will be optimized.
            filter_fun (callable): This must be a function of the model and it must returns the set
              of parameters that the optimizer will consider.
            override_kwargs (dict): The optimizer's keyword arguments to override the default ones.

        Returns:
            tuple[Optimizer, StepLR]: The optimizer and the scheduler.
        """
        if override_kwargs:
            self.optimizer_cfg.update(**override_kwargs)

        if filter_fun is None:
            optimizer = self.optimizer(filter(lambda p: p.requires_grad, model.parameters()),
                                       **self.optimizer_cfg)
        else:
            optimizer = self.optimizer(filter_fun(model),
                                       **self.optimizer_cfg)
        scheduler = self.scheduler(optimizer, **self.scheduler_cfg)
        return optimizer, scheduler

    def __str__(self) -> str:
        strsched = self.scheduler.__name__
        to_str = f"OptCfg({self.optimizer.__name__}, "
        to_str += ", ".join([f"{k}={v}" for k, v in self.optimizer_cfg.items()])
        to_str += f", {strsched}(" + ", ".join([f"{k}={v}" for k, v in self.scheduler_cfg.items()])
        to_str += "))"
        return to_str

    def __repr__(self) -> str:
        return str(self)

    def __getstate__(self) -> dict:
        return self.__dict__

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)


def import_module_from_str(name: str) -> Any:
    """Import a module from its name.

    Args:
        name (str): The name of the module.

    Returns:
        Any: The module.
    """
    components = name.split('.')
    mod = importlib.import_module(".".join(components[:-1]))
    mod = getattr(mod, components[-1])
    return mod


def get_class_from_str(module_name: str, class_name: str) -> Any:
    """Get a class from its name.
    This function is used to get a class from its name and the name of the module where it is
    defined. It is used to dynamically import classes.

    Args:
        module_name (str): The name of the module where the class is defined.
        class_name (str): The name of the class.

    Returns:
        Any: The class.
    """
    module = importlib.import_module(module_name)
    class_ = getattr(module, class_name)
    return class_


def get_class_from_qualified_name(qualname: str) -> Any:
    """Get a class from its fully qualified name.

    Args:
        qualname (str): The fully qualified name of the class.

    Returns:
        Any: The class.
    """
    module_name = ".".join(qualname.split(".")[:-1])
    class_name = qualname.split(".")[-1]
    module = importlib.import_module(module_name)
    class_ = getattr(module, class_name)
    return class_


def get_loss(lname: str) -> Module:
    """Get a loss function from its name.
    The supported loss functions are the ones defined in the ``torch.nn`` module.

    Args:
        lname (str): The name of the loss function.

    Returns:
        Module: The loss function.
    """
    return get_class_from_str("torch.nn", lname)()


def get_model(mname: str, **kwargs: dict[str, Any]) -> Module:
    """Get a model from its name.
    This function is used to get a torch model from its name and the name of the module where it is
    defined. It is used to dynamically import models. If ``mname`` is not a fully qualified name,
    the model is assumed to be defined in the :mod:`fluke.nets` module.

    Args:
        mname (str): The name of the model.
        **kwargs: The keyword arguments to pass to the model's constructor.

    Returns:
        Module: The model.
    """
    if "." in mname:
        module_name = ".".join(mname.split(".")[:-1])
        mname = mname.split(".")[-1]
    else:
        module_name: str = "fluke.nets"
    return get_class_from_str(module_name, mname)(**kwargs)


def get_full_classname(classtype: type) -> str:
    """Get the fully qualified name of a class.

    Args:
        classtype (type): The class.

    Returns:
        str: The fully qualified name of the class.

    Example:
        Let ``A`` be a class defined in the module ``fluke.utils``

        .. code-block:: python
            :linenos:

            # This is the content of the file fluke/utils.py
            class A:
                pass

            get_full_classname(A) # 'fluke.utils.A'

        If the class is defined in the ``__main__`` module, then:

        .. code-block:: python
            :linenos:

            if __name__ == "__main__":
                class B:
                    pass

                get_full_classname(B) # '__main__.B'
    """
    return f"{classtype.__module__}.{classtype.__name__}"


def get_optimizer(oname: str) -> type[Optimizer]:
    """Get an optimizer from its name.
    This function is used to get an optimizer from its name. It is used to dynamically import
    optimizers. The supported optimizers are the ones defined in the :mod:`torch.optim` module.

    Args:
        oname (str): The name of the optimizer.

    Returns:
        type[Optimizer]: The optimizer class.
    """
    return get_class_from_str("torch.optim", oname)


def get_scheduler(sname: str) -> type[LRScheduler]:
    """Get a learning rate scheduler from its name.
    This function is used to get a learning rate scheduler from its name. It is used to dynamically
    import learning rate schedulers. The supported schedulers are the ones defined in the
    :mod:`torch.optim.lr_scheduler` module.

    Args:
        sname (str): The name of the scheduler.

    Returns:
        torch.nn.Module: The learning rate scheduler.
    """
    return get_class_from_str("torch.optim.lr_scheduler", sname)


def clear_cuda_cache(ipc: bool = False):
    """Clear the CUDA cache. This function should be used to free the GPU memory after the training
    process has ended. It is usually used after the local training of the clients.

    Args:
        ipc (bool, optional): Whether to force collecting GPU memory after it has been released by
            CUDA IPC.
    """
    torch.cuda.empty_cache()
    if ipc and torch.cuda.is_available():
        torch.cuda.ipc_collect()


class Configuration(DDict):
    """Fluke configuration class.
    This class is used to store the configuration of an experiment. The configuration must adhere to
    a specific structure. The configuration is validated when the class is instantiated.

    Args:
        config_exp_path (str): The path to the experiment configuration file.
        config_alg_path (str): The path to the algorithm configuration file.
        force_validation (bool, optional): Whether to force the validation of the configuration.
            Defaults to ``True``.

    Raises:
        ValueError: If the configuration is not valid.
    """

    def __init__(self,
                 config_exp_path: str = None,
                 config_alg_path: str = None,
                 force_validation: bool = True):

        if config_exp_path is not None and os.path.exists(config_exp_path):
            with open(config_exp_path) as f:
                config_exp = yaml.safe_load(f)
            self.update(**config_exp)

        if config_alg_path is not None and os.path.exists(config_alg_path):
            with open(config_alg_path) as f:
                config_alg = yaml.safe_load(f)
            self.update(method=config_alg)

        if force_validation:
            self._validate()

    @classmethod
    def from_ddict(cls, ddict: DDict) -> Configuration:
        """Create a configuration from a :class:`fluke.DDict`.

        Args:
            ddict (DDict): The dictionary.

        Returns:
            Configuration: The configuration.
        """
        cfg = Configuration("", "", force_validation=False)
        cfg.update(**ddict)
        cfg._validate()
        return cfg

    def to_dict(self) -> dict:
        """Convert the configuration to a dictionary.

        Returns:
            dict: The dictionary.
        """
        def _to_dict(ddict: DDict) -> dict:
            if not isinstance(ddict, dict):
                if isinstance(ddict, type):
                    return ddict.__name__
                return ddict
            return {k: _to_dict(v) for k, v in ddict.items()}

        return _to_dict(self)

    @classmethod
    def sweep(cls,
              config_exp_path: str,
              config_alg_path: str) -> Generator[Configuration, None, None]:
        """Generate configurations from a sweep.
        This method is used to generate configurations from a sweep. The sweep is defined by the
        experiment configuration file. The method yields a configuration for each combination of
        hyperparameters.

        Args:
            config_exp_path (str): The path to the experiment configuration file.
            config_alg_path (str): The path to the algorithm configuration file.

        Yields:
            Configuration: A configuration.
        """
        cfgs = Configuration(config_exp_path, config_alg_path)

        for cfg in Configuration.__sweep(cfgs):
            yield Configuration.from_ddict(cfg)

    @staticmethod
    def __sweep(cfgs: DDict) -> Generator[DDict, None, None]:
        """Generate configurations from a sweep.
        This method is used to generate configurations from a sweep. The sweep is defined by the
        experiment configuration file. The method yields a configuration for each combination of
        hyperparameters.

        Args:
            cfgs (DDict): The configuration.

        Yields:
            DDict: A configuration.
        """
        if not isinstance(cfgs, dict):
            yield cfgs
            return

        if not cfgs:
            yield {}
            return

        keys, values = zip(*[
            (k, (Configuration.__sweep(v) if isinstance(v, dict)
                 else (v if isinstance(v, list) else [v])))
            for k, v in cfgs.items()
        ])

        for combination in product(*values):
            yield dict(zip(keys, combination))

    @property
    def client(self) -> DDict:
        """Get quick access to the client's hyperparameters.

        Returns:
            DDict: The client's hyperparameters.
        """
        return self.method.hyperparameters.client

    @property
    def server(self) -> DDict:
        """Get quick access to the server's hyperparameters.

        Returns:
            DDict: The server's hyperparameters.
        """
        return self.method.hyperparameters.server

    @property
    def model(self) -> DDict:
        """Get quick access to the model hyperparameters.

        Returns:
            DDict: The model hyperparameters.
        """
        return self.method.hyperparameters.model

    def _validate(self) -> bool:

        EXP_OPT_KEYS = {
            "device": "cpu",
            "seed": 42,
            "inmemory": True
        }

        LOG_OPT_KEYS = {
            "name": "Log"
        }

        SAVE_REQUIRED_KEYS = ["path"]

        SAVE_OPT_KEYS = {
            "save_every": -1,
            "global_only": False
        }

        EVAL_OPT_KEYS = {
            "task": "classification",
            "eval_every": 1,
            "pre_fit": False,
            "post_fit": True,
            "server": True,
            "locals": False
        }

        FIRST_LVL_KEYS = ["data", "protocol", "method"]
        FIRST_LVL_OPT_KEYS = {
            "eval": EVAL_OPT_KEYS,
            "exp": EXP_OPT_KEYS,
            "logger": LOG_OPT_KEYS
        }
        PROTO_REQUIRED_KEYS = ["n_clients", "eligible_perc", "n_rounds"]
        DATA_REQUIRED_KEYS = ["distribution", "dataset"]
        DATA_OPT_KEYS = {
            "sampling_perc": 1.0,
            "client_split": 0.0,
            "keep_test": True,
            "server_test": True,
            "server_split": 0.0,
            "uniform_test": False
        }

        ALG_1L_REQUIRED_KEYS = ["name", "hyperparameters"]
        HP_REQUIRED_KEYS = ["model", "client", "server"]
        CLIENT_HP_REQUIRED_KEYS = ["loss", "batch_size", "local_epochs", "optimizer"]
        WANDB_REQUIRED_KEYS = ["project", "entity"]

        error = False
        for k in FIRST_LVL_KEYS:
            if k not in self:
                f = "experiment" if k != "method" else "algorithm"
                rich.print(f"Error: {k} is required in the {f} configuration file")
                error = True

        for k, v in FIRST_LVL_OPT_KEYS.items():
            if k not in self:
                self[k] = DDict(**v)

        for k in PROTO_REQUIRED_KEYS:
            if k not in self.protocol:
                rich.print(f"Error: {k} is required for key 'protocol'.")
                error = True

        for k in DATA_REQUIRED_KEYS:
            if k not in self.data:
                rich.print(f"Error: {k} is required for key 'data'.")
                error = True

        for k, v in DATA_OPT_KEYS.items():
            if k not in self.data:
                self.data[k] = v

        for k, v in EVAL_OPT_KEYS.items():
            if k not in self.eval:
                self.eval[k] = v

        for k, v in EXP_OPT_KEYS.items():
            if k not in self.exp:
                self.exp[k] = v

        for k, v in LOG_OPT_KEYS.items():
            if k not in self.logger:
                self.logger[k] = v

        for k in ALG_1L_REQUIRED_KEYS:
            if k not in self.method:
                rich.print(f"Error: {k} is required for key 'method'.")
                error = True

        for k in HP_REQUIRED_KEYS:
            if k not in self.method.hyperparameters:
                rich.print(f"Error: {k} is required for key 'hyperparameters' in 'method'.")
                error = True

        for k in CLIENT_HP_REQUIRED_KEYS:
            if k not in self.method.hyperparameters.client:
                rich.print(f"Error: {k} is required as hyperparameter of 'client'.")
                error = True

        if "logger" in self and self.logger.name in ["wandb", "WandBLog"]:
            for k in WANDB_REQUIRED_KEYS:
                if k not in self.logger:
                    rich.print(f"Error: {k} is required for key 'logger' when using 'WandBLog'.")
                    error = True

        if "save" in self and self.save:
            for k in SAVE_REQUIRED_KEYS:
                if k not in self.save:
                    rich.print(f"Error: {k} is required for key 'save'.")
                    error = True
            for k in SAVE_OPT_KEYS:
                if k not in self.save:
                    self.save[k] = SAVE_OPT_KEYS[k]
        else:
            self.save = {}

        if error:
            raise ValueError("Configuration validation failed.")

    def __str__(self) -> str:
        return f"{self.method.name if 'method' in self else 'EXP'}" + \
            f"_data({self.data.dataset.name}, " + \
            f"{self.data.distribution.name}(" + \
            ", ".join([f"{k}={v}" for k, v in self.data.distribution.exclude('name').items()]) +\
            f"))_proto(C{self.protocol.n_clients}, R{self.protocol.n_rounds}, " + \
            f"E{self.protocol.eligible_perc})" + \
            f"_seed({self.exp.seed})"

    def __repr__(self) -> str:
        return str(self)

    def verbose(self) -> str:
        return super().__str__()


def plot_distribution(clients: list[Client],
                      train: bool = True,
                      type: str = "ball") -> None:
    """Plot the distribution of classes for each client.
    This function is used to plot the distribution of classes for each client. The plot can be a
    scatter plot, a heatmap, or a bar plot. The scatter plot (``type='ball'``) shows filled circles
    whose size is proportional to the number of examples of a class. The heatmap (``type='mat'``)
    shows a matrix where the rows represent the classes and the columns represent the clients with
    a color intensity proportional to the number of examples of a class. The bar plot
    (``type='bar'``) shows a stacked bar plot where the height of the bars is proportional to the
    number of examples of a class.

    Warning:
        If the number of clients is greater than 30, the type is automatically switched to
        ``'bar'`` for better visualization.

    Args:
        clients (list[Client]): The list of clients.
        train (bool, optional): Whether to plot the distribution on the training set. If ``False``,
            the distribution is plotted on the test set. Defaults to ``True``.
        type (str, optional): The type of plot. It can be ``'ball'``, ``'mat'``, or ``'bar'``.
            Defaults to ``'ball'``.
    """
    assert type in ["bar", "ball", "mat"], "Invalid plot type. Must be 'bar', 'ball' or 'mat'."
    if len(clients) > 30 and type != "bar":
        warnings.warn("Too many clients to plot. Switching to 'bar' plot.")
        type = "bar"

    client = {}
    for c in clients:
        client[c.index] = c.train_set.tensors[1] if train else c.test_set.tensors[1]

    # Count the occurrences of each class for each client
    class_counts = {client_idx: torch.bincount(client_data).tolist()
                    for client_idx, client_data in enumerate(client.values())}

    # Maximum number of classes
    num_classes = max(len(counts) for counts in class_counts.values())

    _, ax = plt.subplots(figsize=(12, 6))
    ax.set_xticks(range(len(client)))

    class_matrix = np.zeros((num_classes, len(client)))
    for client_idx, counts in class_counts.items():
        for class_idx, count in enumerate(counts):
            class_matrix[class_idx, client_idx] = count
            # Adjusting size based on the count
            if type == "ball":
                size = count * 1  # Adjust the scaling factor as needed
                ax.scatter(client_idx, class_idx, s=size, alpha=0.6)
                ax.set_yticks(range(num_classes))
                ax.text(client_idx, class_idx, str(count), va='center',
                        ha='center', color='black', fontsize=9)
    plt.title('Number of Examples per Class for Each Client', fontsize=12)
    ax.grid(False)
    if type == "mat":
        ax.set_yticks(range(num_classes))
        sns.heatmap(class_matrix, ax=ax, cmap="viridis", annot=class_matrix, fmt='g',
                    cbar=False, annot_kws={"fontsize": 6})
    elif type == "bar":
        df = pd.DataFrame(class_matrix.T, index=[f'{i}' for i in range(len(client))],
                          columns=[f'{i}' for i in range(num_classes)])
        step = int(max(1, np.ceil(len(clients) / 50)))
        df.plot(ax=ax,
                kind='bar',
                stacked=True,
                xticks=[x for x in range(0, len(clients), step)],
                color=sns.color_palette('viridis', num_classes)).legend(loc='upper right')
    plt.xlabel('clients')
    plt.ylabel('classes')
    plt.show()


def bytes2human(n: int) -> str:
    """Convert bytes to human-readable format.

    See:
        https://psutil.readthedocs.io/en/latest/#psutil.Process.memory_info

    Args:
        n (int): The number of bytes.

    Returns:
        str: The number of bytes in human-readable format.

    Example:
        .. code-block:: python
            :linenos:

            >>> bytes2human(10000)
            '9.8K'
            >>> bytes2human(100001221)
            '95.4M'
            >>> bytes2human(1024 ** 3)
            '1.0G'

    """
    symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
    prefix = {}
    for i, s in enumerate(symbols):
        prefix[s] = 1 << (i + 1) * 10
    for s in reversed(symbols):
        if abs(n) >= prefix[s]:
            value = float(n) / prefix[s]
            return '%.1f %sB' % (value, s)
    return "%s B" % n


def memory_usage() -> tuple[int, int, int]:
    """Get the memory usage of the current process.

    Returns:
        tuple[int, int, int]: The resident set size (RSS), the virtual memory size (VMS), and the
            current CUDA reserved memory. If the device is not a CUDA device, the reserved memory is
            0.
    """
    proc = psutil.Process(os.getpid())

    if FlukeENV().get_device().type == "cuda":
        cuda_device = torch.device(FlukeENV().get_device())
        current_reserved = torch.cuda.memory_reserved(cuda_device)
    else:
        current_reserved = 0
    return proc.memory_info().rss, proc.memory_info().vms, current_reserved


def retrieve_obj(key: str,
                 party: Client | Server | None = None,
                 pop: bool = True) -> Any:
    """Load an object from the cache (disk).
    If the object is not found in the cache, it returns ``None``.

    Warning:
        This method assumes the cache is already initialized and opened.
        If it is not, it will raise an exception.

    Args:
        key (str, optional): specific key to add to the default filename.
        party (Client | Server, optional): the party for which to load the object. Defaults to
            ``None``.
        pop (bool, optional): Whether to remove the object from the cache after loading it.
            Defaults to ``True``.

    Returns:
        Any: The object retrieved from the cache.
    """
    prefix = ""
    if party is not None:
        prefix = f"c{party.index}_" if hasattr(party, "index") else "server_"
    cache = FlukeENV().get_cache()
    obj = cache.pop(f"{prefix}{key}", copy=(party is None)) if pop else cache.get(f"{prefix}{key}")
    return obj


def cache_obj(obj: Any,
              key: str,
              party: Server | Client | None = None) -> FlukeCache._ObjectRef | None:
    """Move the object from the RAM to the disk cache to free up memory.
    If the object is ``None``, it returns ``None`` without caching it.

    Warning:
        This method assumes the cache is already initialized and opened.
        If it is not, it will raise an exception.

    Args:
        obj (Any): The object to cache.
        key (str, optional): The key associated with the object.
        party (Client | Server): the party for which to unload the model. Defaults to ``None``.

    Returns:
        FlukeCache._ObjectRef: The object reference identifier. If the object is ``None``,
        it returns ``None``.
    """
    if obj is None:
        return None
    prefix = ""
    if party is not None:
        prefix = f"c{party.index}_" if hasattr(party, "index") else "server_"
    cache = FlukeENV().get_cache()
    return cache.push(f"{prefix}{key}", obj)


def _flatten_dict(d: dict, parent_key: str = '', sep: str = '.'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            items.extend(_flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def flatten_dict(nested_dict: dict, sep: str = '.') -> dict:
    """Flatten a nested dictionary.
    The flatten dictionary is a dictionary where the keys are the concatenation of the keys of the
    nested dictionary separated by a separator.

    Args:
        d (dict): Nested dictionary.
        sep (str, optional): Separator. Defaults to '.'.

    Returns:
        dict: Flattened dictionary.

    Example:
        .. code-block:: python
            :linenos:

            d = {'a': 1, 'b': {'c': 2, 'd': {'e': 3}}}
            flatten_dict(d)
            # Output: {'a': 1, 'b.c': 2, 'b.d.e': 3}

    """
    return _flatten_dict(nested_dict)



"""Implementation of the [FedRep21]_ algorithm.

References:
    .. [FedRep21] Liam Collins, Hamed Hassani, Aryan Mokhtari, and Sanjay Shakkottai.
       Exploiting shared representations for personalized federated learning. In ICML (2021).
       URL: https://arxiv.org/abs/2102.07078
"""
import sys
from typing import Any, Iterable
from dataclasses import dataclass
import torch

sys.path.append(".")
sys.path.append("..")

from ..algorithms import CentralizedFL  # NOQA
from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..data import FastDataLoader  # NOQA
from ..nets import EncoderGlobalHeadLocalNet, EncoderHeadNet  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache, get_model  # NOQA
from ..utils.model import safe_load_state_dict, ModOpt  # NOQA

__all__ = [
    "FedRepClient",
    "FedRepServer",
    "FedRep"
]

# https://arxiv.org/abs/2102.07078


@dataclass
class _ModOpt2(ModOpt):

    pers_optimizer: torch.optim.Optimizer = None
    pers_scheduler: torch.optim.lr_scheduler = None


class FedRepClient(Client):

    def __init__(self,
                 index: int,
                 model: EncoderHeadNet,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int = 3,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 tau: int = 3,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, train_set=train_set,
                         test_set=test_set, optimizer_cfg=optimizer_cfg, loss_fn=loss_fn,
                         local_epochs=local_epochs, fine_tuning_epochs=fine_tuning_epochs,
                         clipping=clipping, **kwargs)
        self._load_from_cache()
        if isinstance(model, str):
            model = get_model(model)
        self._modopt = _ModOpt2(model=EncoderGlobalHeadLocalNet(model))
        self.hyper_params.update(tau=tau)
        self._save_to_cache()

    @property
    def pers_optimizer(self) -> torch.optim.Optimizer:
        """Optimizers for the personalized part of the model.

        Returns:
            torch.optim.Optimizer: Optimizer for the personalized part of the model.
        """
        return self._modopt.pers_optimizer

    @pers_optimizer.setter
    def pers_optimizer(self, optimizer: torch.optim.Optimizer) -> None:
        self._modopt.pers_optimizer = optimizer

    @property
    def pers_scheduler(self) -> torch.optim.lr_scheduler:
        """Learning rate scheduler for the personalized part of the model.

        Returns:
            torch.optim.lr_scheduler: Learning rate scheduler for the personalized part of the
                model.
        """
        return self._modopt.pers_scheduler

    @pers_scheduler.setter
    def pers_scheduler(self, scheduler: torch.optim.lr_scheduler) -> None:
        self._modopt.pers_scheduler = scheduler

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)
        self.model.train()
        self.model.to(self.device)

        # update head layers
        for parameter in self.model.get_local().parameters():
            parameter.requires_grad = True
        for parameter in self.model.get_global().parameters():
            parameter.requires_grad = False

        if self.pers_optimizer is None:
            self.pers_optimizer, self.pers_scheduler = self._optimizer_cfg(self.model.get_local())

        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.pers_optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()
                self._clip_grads(self.model)
                self.pers_optimizer.step()
                running_loss += loss.item()
            self.pers_scheduler.step()

        # update encoder layers
        for parameter in self.model.get_local().parameters():
            parameter.requires_grad = False
        for parameter in self.model.get_global().parameters():
            parameter.requires_grad = True

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model.get_global())

        for _ in range(self.hyper_params.tau):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()
                self.optimizer.step()
            self.scheduler.step()

        running_loss /= (epochs * len(self.train_set))
        self.model.cpu()
        clear_cuda_cache()
        return running_loss

    def send_model(self):
        self.channel.send(Message(self.model.get_global(), "model", self, inmemory=True),
                          self.server)

    def receive_model(self) -> None:
        server_model = self.channel.receive(self, self.server, msg_type="model").payload
        if self.model is None:
            self.model = EncoderGlobalHeadLocalNet(server_model)
        safe_load_state_dict(self.model.get_global(), server_model.state_dict())


class FedRepServer(Server):

    def __init__(self,
                 model: torch.nn.Module,
                 test_set: FastDataLoader,  # test_set is not used
                 clients: Iterable[Client],
                 weighted: bool = False):
        super().__init__(model=model, test_set=None, clients=clients, weighted=weighted)


class FedRep(CentralizedFL):

    def get_client_class(self) -> Client:
        return FedRepClient

    def get_server_class(self) -> Server:
        return FedRepServer



from __future__ import annotations

import sys

# import numpy as np
import pytest
import torch
from torchvision.transforms import v2

sys.path.append(".")
sys.path.append("..")

from fluke import DDict  # NOQA
from fluke.data import (DataContainer, DataSplitter, DummyDataContainer,  # NOQA
                        FastDataLoader)
from fluke.data.datasets import Datasets  # NOQA


def test_container():
    data = DataContainer(
        X_train=torch.rand(100, 20),
        y_train=torch.randint(0, 10, (100,)),
        X_test=torch.rand(100, 20),
        y_test=torch.randint(0, 10, (100,)),
        num_classes=10
    )

    assert data.train[0].shape == torch.Size([100, 20])
    assert data.train[1].shape == torch.Size([100])
    assert data.test[0].shape == torch.Size([100, 20])
    assert data.test[1].shape == torch.Size([100])
    assert data.num_classes == 10
    assert data.num_features == 20

    # data.standardize()

    # assert torch.allclose(data.train[0].mean(dim=0), torch.zeros(20), atol=1e-7)
    # assert torch.allclose(data.train[0].std(dim=0), torch.ones(20), atol=1e-2)


def test_ftdl():
    X, y = torch.rand(100, 20), torch.randint(0, 10, (100,))
    loader = FastDataLoader(X,
                            y,
                            num_labels=10,
                            batch_size=10,
                            shuffle=True,
                            percentage=1,
                            skip_singleton=True)

    assert len(loader) == 10
    assert torch.allclose(loader[0][0], X[0])
    assert torch.allclose(loader[0][1], y[0])

    with pytest.raises(IndexError):
        loader[300]

    for X, y in loader:
        assert X.shape == torch.Size([10, 20])
        assert y.shape == torch.Size([10])
        break

    loader.batch_size = 99

    cnt = 0
    for X, y in loader:
        cnt += 1
    assert cnt == 1

    loader.skip_singleton = False

    cnt = 0
    for X, y in loader:
        cnt += 1
    assert cnt == 2

    assert loader.size == 100
    loader.set_sample_size(0.2)
    assert loader.size == 20

    with pytest.raises(ValueError):
        loader.batch_size = -10

    with pytest.raises(ValueError):
        loader.set_sample_size(-10)

    with pytest.raises(ValueError):
        loader.set_sample_size(1.1)

    loader = FastDataLoader(torch.FloatTensor([[1, 2, 3], [4, 5, 6]]),
                            torch.LongTensor([0, 1]),
                            num_labels=10,
                            batch_size=1,
                            shuffle=False,
                            percentage=1,
                            skip_singleton=True)

    for X, y in loader:
        assert X == torch.FloatTensor([[1, 2, 3]])
        assert y == torch.tensor([0])

    loader.shuffle = True

    for X, y in loader:
        assert X == torch.FloatTensor([[4, 5, 6]]) or X == torch.FloatTensor([[1, 2, 3]])
        assert y == torch.tensor([1]) or y == torch.tensor([0])

    loader = FastDataLoader(torch.FloatTensor([[1, 2, 3], [4, 5, 6]]),
                            torch.LongTensor([0, 1]),
                            num_labels=10,
                            batch_size=1,
                            shuffle=False,
                            percentage=1,
                            skip_singleton=False,
                            single_batch=True)

    cnt = 0
    for X, y in loader:
        cnt += 1
    assert cnt == 1

    with pytest.raises(AssertionError):
        loader = FastDataLoader(torch.rand(100, 20),
                                torch.randint(0, 10, (101,)),
                                num_labels=10,
                                batch_size=10,
                                shuffle=True,
                                percentage=1,
                                skip_singleton=True)

    loader = FastDataLoader(torch.FloatTensor([[1, 2, 3], [4, 5, 6]]),
                            torch.LongTensor([0, 1]),
                            num_labels=10,
                            batch_size=0,
                            shuffle=False,
                            percentage=1,
                            skip_singleton=False,
                            single_batch=False)

    assert loader.batch_size == 2

    imgs = torch.randint(0, 256, size=(5, 3, 32, 32), dtype=torch.uint8)
    lbls = torch.randint(0, 10, size=(5,), dtype=torch.long)

    transforms = v2.Compose([
        v2.RandomResizedCrop(size=(224, 224), antialias=False),
        v2.RandomHorizontalFlip(p=0.5),
        v2.ToDtype(torch.float32, scale=True),
        v2.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ])

    loader = FastDataLoader(imgs,
                            lbls,
                            num_labels=10,
                            batch_size=1,
                            shuffle=False,
                            transforms=transforms,
                            percentage=1,
                            skip_singleton=False,
                            single_batch=False)

    for X, y in loader:
        assert X.shape == torch.Size([1, 3, 224, 224])
        assert y.shape == torch.Size([1])
        break

    x0_tr = loader[0]
    assert x0_tr[0].shape == torch.Size([3, 224, 224])
    assert torch.any(x0_tr[0] != X[0])


def test_splitter():
    cfg = DDict(
        client_split=0.1,
        dataset={
            "name": "mnist",
        },
        distribution={
            "name": "iid",
        },
        sampling_perc=0.1,
        server_test=True,
        server_split=0.2,
        keep_test=False
    )

    data_container = Datasets.get(**cfg.dataset)
    splitter = DataSplitter(dataset=data_container,
                            distribution=cfg.distribution.name,
                            dist_args=cfg.distribution.exclude("name"),
                            **cfg.exclude('dataset', 'distribution'))
    assert splitter.client_split == 0.1
    assert splitter.sampling_perc == 0.1
    assert splitter.server_split == 0.2
    assert not splitter.keep_test
    assert splitter.distribution == "iid"

    # OK uniform
    (ctr, cte), ste = splitter.assign(10, batch_size=10)

    assert len(ctr) == 10
    assert len(cte) == 10
    assert isinstance(ctr[0], FastDataLoader)
    x, y = next(iter(ctr[0]))
    assert x.shape == torch.Size([10, 28, 28])
    assert y.shape == torch.Size([10])
    x, y = next(iter(cte[0]))
    assert x.shape == torch.Size([10, 28, 28])
    assert y.shape == torch.Size([10])
    assert isinstance(ste, FastDataLoader)
    x, y = next(iter(ste))
    assert x.shape == torch.Size([128, 28, 28])
    assert y.shape == torch.Size([128])

    n_clients = 100

    # OK
    splitter.distribution = "dir"
    (ctr, cte), ste = splitter.assign(n_clients, batch_size=10)

    # OK?
    # splitter.distribution = "covariate"
    # (ctr, cte), ste = splitter.assign(n_clients, batch_size=10)

    # OK
    # splitter.distribution = "classwise_qnt"
    # (ctr, cte), ste = splitter.assign(n_clients, batch_size=10)

    splitter.distribution = "pathological"
    (ctr, cte), ste = splitter.assign(n_clients, batch_size=10)

    splitter.distribution = "lbl_qnt"
    (ctr, cte), ste = splitter.assign(n_clients, batch_size=10)

    splitter.distribution = "qnt"
    (ctr, cte), ste = splitter.assign(n_clients, batch_size=10)

    splitter.distribution = "dir"
    splitter.dist_args.balanced = False
    (ctr, cte), ste = splitter.assign(n_clients, batch_size=10)

    # freq = []
    # for i in range(n_clients):
    #     arr = np.concatenate([ctr[i].tensors[1].numpy(), cte[i].tensors[1].numpy()])
    #     freq.append({val: np.sum(arr == val) for val in np.unique(arr)})
    #     # print(i, freq[-1])

    # freq0 = np.array([freq[i][0] - 2 for i in range(n_clients)], dtype=float)
    # freq0 /= np.sum(freq0, dtype=float)

    # print(",".join([str(x) for x in freq0]))

    # keep_test
    # - server_test + client_split = 0
    # - server_test + client_split > 0
    # - not server_test + client_split > 0
    # not keep_test
    # - server_test + client_split = 0
    # - server_test + client_split > 0
    # - not server_test + client_split > 0

    # ERROR: keep_test=False, server_test=False, client_split=0
    # ERROR: keep_test=False, server_test=True, server_split=0
    cfg = DDict(
        dataset={
            "name": "mnist",
        },
        distribution={
            "name": "iid",
        },
        sampling_perc=0.1,
        keep_test=True,
        server_test=True,
        server_split=0.2,
        client_split=0.1,
        uniform_test=True
    )
    data_container = Datasets.get(**cfg.dataset)
    splitter = DataSplitter(dataset=data_container,
                            distribution=cfg.distribution.name,
                            dist_args=cfg.distribution.exclude("name"),
                            **cfg.exclude('dataset', 'distribution'))
    (ctr, cte), ste = splitter.assign(n_clients=10, batch_size=10)

    n_examples = splitter.data_container.train[0].shape[0]
    assert ctr[0].size == int(n_examples / 10 * 0.1 * 0.9)
    assert cte[0].size == int(n_examples / 10 * 0.1 * 0.1)
    assert ste.size == 1000

    # cfg.client_split = 0.2
    # splitter = DataSplitter(dataset=data_container,
    #                         distribution="dir",
    #                         dist_args={"beta": 0.3},
    #                         **cfg.exclude('dataset', 'distribution'))

    # (ctr, cte), ste = splitter.assign(n_clients=500, batch_size=10)

    cfg.client_split = 0
    splitter = DataSplitter(dataset=data_container,
                            distribution=cfg.distribution.name,
                            dist_args=cfg.distribution.exclude("name"),
                            **cfg.exclude('dataset', 'distribution'))
    (ctr, cte), ste = splitter.assign(n_clients=10, batch_size=10)

    assert ctr[0].size == int(n_examples / 10 * 0.1)
    assert cte[0] is None
    assert ste.size == 1000

    cfg.server_test = False
    cfg.client_split = 0.1
    splitter = DataSplitter(dataset=data_container,
                            distribution=cfg.distribution.name,
                            dist_args=cfg.distribution.exclude("name"),
                            **cfg.exclude('dataset', 'distribution'))
    (ctr, cte), ste = splitter.assign(n_clients=10, batch_size=10)

    n_examples_te = splitter.data_container.test[0].shape[0]
    assert ctr[0].size == int(n_examples / 10 * 0.1)
    assert cte[0].size == int(n_examples_te / 10 * 0.1)
    assert ste is None

    cfg.keep_test = False
    splitter = DataSplitter(dataset=data_container,
                            distribution=cfg.distribution.name,
                            dist_args=cfg.distribution.exclude("name"),
                            **cfg.exclude('dataset', 'distribution'))
    (ctr, cte), ste = splitter.assign(n_clients=10, batch_size=10)

    assert ctr[0].size == int((n_examples + n_examples_te) / 10 * 0.1 * 0.9)
    assert cte[0].size == int((n_examples + n_examples_te) / 10 * 0.1 * 0.1)
    assert ste is None

    cfg.client_split = 0
    cfg.server_test = True
    cfg.server_split = 0.2
    splitter = DataSplitter(dataset=data_container,
                            distribution=cfg.distribution.name,
                            dist_args=cfg.distribution.exclude("name"),
                            **cfg.exclude('dataset', 'distribution'))
    (ctr, cte), ste = splitter.assign(n_clients=10, batch_size=10)

    assert ctr[0].size == int((n_examples + n_examples_te) * 0.8 / 10 * 0.1)
    assert cte[0] is None
    assert ste.size == int((n_examples + n_examples_te) * 0.1 * 0.2)

    cfg.client_split = 0
    cfg.keep_test = False
    cfg.server_test = False
    with pytest.raises(AssertionError):
        DataSplitter(data_container, **cfg.exclude("dataset"))

    cfg.server_test = True
    cfg.server_split = 0
    with pytest.raises(AssertionError):
        DataSplitter(data_container, **cfg.exclude("dataset"))

    dummy = DummyDataContainer(ctr, cte, ste, 10)
    assert dummy.num_classes == 10
    assert dummy.clients_tr == ctr
    assert dummy.clients_te == cte
    assert dummy.server_data == ste

    splitter = DataSplitter(dataset=dummy)
    (ctr_, cte_), ste_ = splitter.assign(10, batch_size=10)
    assert torch.all(ctr_[0][0][0] == ctr[0][0][0])
    assert cte_[0] is None
    assert torch.all(ste_[0][0][0] == ste[0][0][0])


if __name__ == "__main__":
    test_container()
    test_ftdl()
    test_splitter()

    # 95% coverage on fluke/data



"""
The module :mod:`fluke.server` provides the base classes for the servers in :mod:`fluke`.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generator, Iterable, Union

import numpy as np
import torch
from rich.progress import track

from . import DDict, FlukeENV, ObserverSubject  # NOQA
from .comm import Channel, Message  # NOQA
from .data import FastDataLoader  # NOQA
from .evaluation import Evaluator  # NOQA
from .utils.model import aggregate_models  # NOQA

if TYPE_CHECKING:
    from .client import Client


__all__ = [
    "Server"
]

torch.serialization.add_safe_globals([set])


class Server(ObserverSubject):
    """Basic Server for Federated Learning.

    This class is the base class for all servers in :mod:`fluke`. It implements the basic
    functionalities of a federated learning server. The default behaviour of this server is based
    on the Federated Averaging algorithm. The server is responsible for coordinating the learning
    process, selecting the clients for each round, sending the global model to the clients, and
    aggregating the models received from the clients at the end of the round. The server also
    evaluates the model server-side (if the test data is provided).

    Attributes:
        hyper_params (DDict):
          The hyper-parameters of the server. The default hyper-parameters are:

          - weighted: A boolean indicating if the clients should be weighted by the number of
            samples when aggregating the models.

          When a new server class inherits from this class, it must add all its hyper-parameters
          to this dictionary.
        device (torch.device): The device where the server runs.
        model (torch.nn.Module): The federated model to be trained.
        clients (Iterable[Client]): The clients that will participate in the federated learning
          process.
        n_clients (int): The number of clients that will participate in the federated learning
        rounds (int): The number of rounds that have been executed.
        test_set (FastDataLoader): The test data to evaluate the model. If None, the model
          will not be evaluated server-side.

    Args:
        model (torch.nn.Module): The federated model to be trained.
        test_set (FastDataLoader): The test data to evaluate the model.
        clients (Iterable[Client]): The clients that will participate in the federated learning
          process.
        weighted (bool): A boolean indicating if the clients should be weighted by the
          number of samples when aggregating the models. Defaults to False.
    """

    def __init__(self,
                 model: torch.nn.Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 weighted: bool = False,
                 lr: float = 1.0,
                 **kwargs: dict[str, Any]):
        super().__init__()
        self.hyper_params = DDict(
            weighted=weighted,
            lr=lr
        )
        self.device: torch.device = FlukeENV().get_device()
        self.model: torch.nn.Module = model
        self.clients: Iterable[Client] = clients
        self.n_clients: int = len(clients)
        self.rounds: int = 0
        self.test_set: FastDataLoader = test_set

        self._channel: Channel = Channel()
        self._participants: set[int] = set()

    @property
    def channel(self) -> Channel:
        """The channel to communicate with the clients.

        Important:
            Always use this channel to exchange data/information with the clients.
            The server should directly call the clients' methods only to trigger specific actions.

        Returns:
            Channel: The channel to communicate with the clients.
        """
        return self._channel

    @property
    def has_test(self) -> bool:
        """Return whether the server can evaluate the model.

        Returns:
            bool: True if the server can evaluate the model, False otherwise.
        """
        return self.test_set is not None

    @property
    def has_model(self) -> bool:
        """Return whether the server owns a global model.

        Returns:
            bool: True if the server owns a global model, False otherwise.
        """
        return self.model is not None

    def broadcast_model(self, eligible: Iterable[Client]) -> None:
        """Broadcast the global model to the clients.

        Args:
            eligible (Iterable[Client]): The clients that will receive the global model.
        """
        self.channel.broadcast(Message(self.model, "model", self, inmemory=None), eligible)

    def fit(self,
            n_rounds: int = 10,
            eligible_perc: float = 0.1,
            finalize: bool = True,
            **kwargs: dict[str, Any]) -> None:
        """Run the federated learning algorithm.
        The default behaviour of this method is to run the Federated Averaging algorithm. The server
        selects a percentage of the clients to participate in each round, sends the global model to
        the clients, which compute the local updates and send them back to the server. The server
        aggregates the models of the clients and repeats the process for a number of rounds.

        Args:
            n_rounds (int, optional): The number of rounds to run. Defaults to 10.
            eligible_perc (float, optional): The percentage of clients that will be selected for
              each round. Defaults to 0.1.
            finalize (bool, optional): If True, the server will finalize the federated learning
                process. Defaults to True.
            **kwargs: Additional keyword arguments.
        """
        with FlukeENV().get_live_renderer():
            progress_fl = FlukeENV().get_progress_bar("FL")
            progress_client = FlukeENV().get_progress_bar("clients")
            client_x_round = int(self.n_clients * eligible_perc)
            task_rounds = progress_fl.add_task("[red]FL Rounds", total=n_rounds*client_x_round)
            task_local = progress_client.add_task("[green]Local Training", total=client_x_round)

            total_rounds = self.rounds + n_rounds
            for round in range(self.rounds, total_rounds):
                try:
                    self._notify_start_round(round + 1, self.model)
                    eligible = self.get_eligible_clients(eligible_perc)
                    self._notify_selected_clients(round + 1, eligible)
                    self.broadcast_model(eligible)

                    for c, client in enumerate(eligible):
                        client.local_update(round + 1)
                        self._participants.update([client.index])
                        progress_client.update(task_id=task_local, completed=c+1)
                        progress_fl.update(task_id=task_rounds, advance=1)

                    client_models = self.receive_client_models(eligible, state_dict=False)
                    self.aggregate(eligible, client_models)
                    self._compute_evaluation(round, eligible)
                    self._notify_end_round(round + 1)
                    self.rounds += 1

                except KeyboardInterrupt:
                    self._notify_interrupt()
                    break
            progress_fl.remove_task(task_rounds)
            progress_client.remove_task(task_local)

        if finalize:
            self.finalize()

    def _compute_evaluation(self, round: int, eligible: Iterable[Client]) -> None:
        evaluator = FlukeENV().get_evaluator()

        if FlukeENV().get_eval_cfg().locals:
            client_evals = {client.index: client.evaluate(evaluator, self.test_set)
                            for client in eligible}
            self._notify_evaluation(round + 1, "locals", client_evals)

        if FlukeENV().get_eval_cfg().server:
            evals = self.evaluate(evaluator, self.test_set)
            self._notify_evaluation(round + 1, "global", evals)

    def evaluate(self, evaluator: Evaluator, test_set: FastDataLoader) -> dict[str, float]:
        """Evaluate the global federated model on the :attr:`test_set`.
        If the test set is not set, the method returns an empty dictionary.

        Returns:
            dict[str, float]: The evaluation results. The keys are the metrics and the values are
                the results.
        """
        if test_set is not None:
            return evaluator.evaluate(self.rounds + 1, self.model, test_set, device=self.device)
        return {}

    def finalize(self) -> None:
        """Finalize the federated learning process.
        The finalize method is called at the end of the federated learning process. The client-side
        evaluation is only done if the client has participated in at least one round.
        """
        if self.rounds == 0:
            return
        client_to_eval = [client for client in self.clients if client.index in self._participants]
        self.broadcast_model(client_to_eval)
        for client in track(client_to_eval, "Finalizing federation...", transient=True):
            client.finalize()
        # self._compute_evaluation(self.rounds, client_to_eval)
        self._notify_finalize()

    def get_eligible_clients(self, eligible_perc: float) -> Iterable[Client]:
        """Get the clients that will participate in the current round.
        Clients are selected randomly based on the percentage of eligible clients.

        Args:
            eligible_perc (float): The percentage of clients that will be selected.

        Returns:
            Iterable[Client]: The clients that will participate in the current round.
        """
        if eligible_perc == 1:
            if not self._participants:
                self._participants = set(range(self.n_clients))
            return self.clients
        n = int(self.n_clients * eligible_perc)
        selected = np.random.choice(self.clients, n, replace=False)
        return selected

    def receive_client_models(self,
                              eligible: Iterable[Client],
                              state_dict: bool = True) -> Generator[torch.nn.Module, None, None]:
        """Retrieve the models of the clients.
        This method assumes that the clients have already sent their models to the server.
        The models are received through the channel in the same order as the clients in
        ``eligible``.

        Caution:
            The method returns a generator of the models of the clients to avoid to clutter the
            memory with all the models. This means that this method is expected to be called only
            once per round. If the models are needed multiple times, the generator should be
            converted to a list, tuple, or any other iterable.

        Args:
            eligible (Iterable[Client]): The clients that will participate in the aggregation.
            state_dict (bool, optional): If True, the method returns the state_dict of the models.
              Otherwise, it returns the models. Defaults to True.

        Returns:
            Generator[torch.nn.Module]: The models of the clients.
        """
        for client in eligible:
            client_model = self.channel.receive(self, client, "model").payload
            if state_dict:
                client_model = client_model.state_dict()
            yield client_model

    def _get_client_weights(self, eligible: Iterable[Client]) -> list[float]:
        """Get the weights of the clients for the aggregation.
        The weights are calculated based on the number of samples of each client.
        If the hyperparameter ``weighted`` is True, the clients are weighted by their number of
        samples. Otherwise, all clients have the same weight.

        Caution:
            The computation of the weights do not adhere to the "best-practices" of :mod:`fluke`
            because the server should not have direct access to the number of samples of the
            clients. Thus, the computation of the weights should be done communicating with the
            clients through the channel but, for simplicity, we are not following this practice
            here. However, the communication overhead is negligible and does not affect the logged
            performance.

        Args:
            eligible (Iterable[Client]): The clients that will participate in the aggregation.

        Returns:
            list[float]: The weights of the clients.
        """
        if "weighted" in self.hyper_params.keys() and self.hyper_params.weighted:
            num_ex = [client.n_examples for client in eligible]
            tot_ex = sum(num_ex)
            return [num_ex[i] / tot_ex for i in range(len(eligible))]
        else:
            return [1. / len(eligible)] * len(eligible)

    @torch.no_grad()
    def aggregate(self,
                  eligible: Iterable[Client],
                  client_models: Iterable[torch.nn.Module]) -> None:
        r"""Aggregate the models of the clients.
        The aggregation is done by averaging the models of the clients. If the hyperparameter
        ``weighted`` is ``True``, the clients are weighted by their number of samples.
        The method directly updates the model of the server. Formally, let :math:`\theta` be the
        model of the server, :math:`\theta_i` the model of client :math:`i`, and :math:`w_i` the
        weight of client :math:`i` such that :math:`\sum_{i=1}^{N} w_i = 1`. The aggregation is
        done as follows [FedAVG]_:

        .. math::
            \theta = \sum_{i=1}^{N} w_i \theta_i

        Note:
            In case of networks with batch normalization layers, the running statistics of the
            batch normalization layers are also aggregated. For all statistics but
            ``num_batches_tracked`` are aggregated the mean is computed, while for the
            ``num_batches_tracked`` parameter, the maximum between the clients' values is taken.

        See also:
            :func:`fluke.utils.model.aggregate_models`

        Args:
            eligible (Iterable[Client]): The clients that will participate in the aggregation.
            client_models (Iterable[torch.nn.Module]): The models of the clients.

        References:
            .. [FedAVG] H. B. McMahan, E. Moore, D. Ramage, S. Hampson, and B. A. y Arcas,
               "Communication-Efficient Learning of Deep Networks from Decentralized Data".
               In AISTATS (2017).
        """
        # agg_model_sd = OrderedDict()
        weights = self._get_client_weights(eligible)
        aggregate_models(self.model, client_models, weights, self.hyper_params.lr, inplace=True)

    def _notify_start_round(self, round: int, global_model: Any) -> None:
        """Notify the observers that a new round has started.

        Args:
            round (int): The round number.
            global_model (Any): The current global model.
        """
        for observer in self._observers:
            observer.start_round(round, global_model)

    def _notify_evaluation(self,
                           round: int,
                           type: str,
                           evals: Union[dict[str, float], dict[int, dict[str, float]]]) -> None:
        """Notify the observers that a round has ended.

        Args:
            round (int): The round number.
            global_model (Any): The current global model.
            data (FastDataLoader): The test data.
            client_evals (Iterable[Any]): The evaluation metrics of the clients.
        """
        for observer in self._observers:
            observer.server_evaluation(round, type, evals)

    def _notify_end_round(self, round: int) -> None:
        """Notify the observers that a round has ended.

        Args:
            round (int): The round number.
            global_model (Any): The current global model.
            data (FastDataLoader): The test data.
            client_evals (Iterable[Any]): The evaluation metrics of the clients.
        """
        for observer in self._observers:
            observer.end_round(round)

    def _notify_selected_clients(self, round: int, clients: Iterable[Any]) -> None:
        """Notify the observers that the clients have been selected for the current round.

        Args:
            round (int): The round number.
            clients (Iterable[Any]): The clients selected for the current round.
        """
        for observer in self._observers:
            observer.selected_clients(round, clients)

    def _notify_finalize(self) -> None:
        """Notify the observers that the federated learning process has ended.

        Args:
            evals (dict[str, float]): The evaluation metrics of the global model.
            client_evals (Iterable[Any]): The evaluation metrics of the clients.
        """
        for observer in self._observers:
            observer.finished(self.rounds + 1)

    def _notify_interrupt(self) -> None:
        """Notify the observers that the federated learning process has been interrupted by
        the user.
        """
        for observer in self._observers:
            observer.interrupted()

    def __str__(self) -> str:
        hpstr = ", ".join([f"{h}={str(v)}" for h, v in self.hyper_params.items()])
        return f"{self.__class__.__name__}({hpstr})"

    def __repr__(self) -> str:
        return str(self)

    def state_dict(self) -> dict:
        """Return the server's state as a dictionary.

        Returns:
            dict: The server's state.
        """
        return {
            "model": self.model.state_dict(),
            "rounds": self.rounds,
            "participants": self._participants
        }

    def save(self, path: str) -> None:
        """Save the server's state to file.

        Args:
            path (str): The path to save the server.
        """
        torch.save(self.state_dict(), path)

    def load(self, path: str) -> None:
        """Load the server's state from file.

        Args:
            path (str): The path to load the server's state.
        """
        state = torch.load(path, weights_only=True)
        self.model.load_state_dict(state["model"])
        self.rounds = state["rounds"]
        self._participants = set(state["participants"])



"""Implementation of the Federated Averaging with Body Aggregation and Body Update [FedBABU22]_
algorithm.

References:
    .. [FedBABU22] Jaehoon Oh, Sangmook Kim, Se-Young Yun. FedBABU: Towards Enhanced Representation
       for Federated Image Classification. In ICLR (2022). URL: https://arxiv.org/abs/2106.06042
"""
import sys
from typing import Any, Iterable

from rich.progress import Progress
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from .. import FlukeENV  # NOQA
from ..algorithms import PersonalizedFL  # NOQA
from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..data import FastDataLoader  # NOQA
from ..nets import EncoderHeadNet  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache  # NOQA
from ..utils.model import safe_load_state_dict  # NOQA


class FedBABUClient(Client):

    def __init__(self,
                 index: int,
                 model: EncoderHeadNet,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: Module,
                 local_epochs: int,
                 mode: str,
                 fine_tuning_epochs: int,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):
        assert mode in ["head", "body", "full"]
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.hyper_params.update(mode=mode)
        self._load_from_cache()
        self.model = model
        self._save_to_cache()

    def send_model(self):
        self.channel.send(Message(self.model.encoder,
                          "model", self, inmemory=True), self.server)

    def receive_model(self) -> None:
        msg = self.channel.receive(self, self.server, msg_type="model")
        safe_load_state_dict(self.model.encoder, msg.payload.state_dict())

        # Deactivate gradient
        for param in self.model.head.parameters():
            param.requires_grad = False

    def fine_tune(self) -> None:
        """Fine-tune the personalized model."""

        self._load_from_cache()
        if self.hyper_params.mode == "full":
            for param in self.model.parameters():
                param.requires_grad = True
        elif self.hyper_params.mode == "body":
            for param in self.model.encoder.parameters():
                param.requires_grad = True
            for param in self.model.head.parameters():
                param.requires_grad = False
        else:  # head
            for param in self.model.encoder.parameters():
                param.requires_grad = False
            for param in self.model.head.parameters():
                param.requires_grad = True

        self.model.train()
        self.model.to(self.device)
        self.pers_optimizer, self.pers_scheduler = self._optimizer_cfg(self.model)

        for _ in range(self.hyper_params.fine_tuning_epochs):
            loss = None
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.pers_optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()
                self.pers_optimizer.step()
            self.pers_scheduler.step()

        self.model.cpu()
        clear_cuda_cache()
        self._save_to_cache()

    def finalize(self) -> None:
        self._load_from_cache()
        metrics = self.evaluate(FlukeENV().get_evaluator(), self.test_set)
        if metrics:
            self._notify_evaluation(-1, "post-fit", metrics)
        self._save_to_cache()


class FedBABUServer(Server):

    def __init__(self,
                 model: Module,
                 test_set: FastDataLoader,  # not used
                 clients: Iterable[Client],
                 weighted: bool = False):
        super().__init__(model=model, test_set=None, clients=clients, weighted=weighted)

    def finalize(self) -> None:

        with Progress(transient=True) as progress:
            task = progress.add_task("[cyan]Client's fine tuning", total=len(self._participants))
            clients_ft = [client for client in self.clients if client.index in self._participants]
            for client in clients_ft:
                client.fine_tune()
                client.finalize()
                progress.update(task, advance=1)

        self._notify_finalize()


class FedBABU(PersonalizedFL):

    def get_client_class(self) -> Client:
        return FedBABUClient

    def get_server_class(self) -> Server:
        return FedBABUServer



"""This submodule provides logging utilities."""
import json
import os
import sys
import time
from typing import Any, Literal, Union

import numpy as np
import pandas as pd
import psutil
import rich
from rich.panel import Panel
from rich.pretty import Pretty
from torch.nn import Module
from torch.utils.tensorboard import SummaryWriter

import wandb

sys.path.append(".")
sys.path.append("..")

from .. import DDict  # NOQA
from ..comm import ChannelObserver, Message  # NOQA
from ..utils import bytes2human  # NOQA
from . import ClientObserver, ServerObserver, get_class_from_str  # NOQA


__all__ = [
    "Log",
    "TensorboardLog",
    "WandBLog",
    "ClearMLLog",
    "get_logger"
]


class Log(ServerObserver, ChannelObserver, ClientObserver):
    """Basic logger.
    This class is used to log the performance of the global model and the communication costs during
    the federated learning process. The logging happens in the console.

    Attributes:
        global_eval (dict): The global evaluation metrics.
        locals_eval (dict): The clients local model evaluation metrics on the server's test set.
        prefit_eval (dict): The clients' pre-fit evaluation metrics.
        postfit_eval (dict): The clients' post-fit evaluation metrics.
        locals_eval_summary (dict): The mean of the clients local model evaluation metrics.
        prefit_eval_summary (dict): The mean of the clients pre-fit evaluation metrics.
        postfit_eval_summary (dict): The mean of the clients post-fit evaluation metrics.
        comm_costs (dict): The communication costs.
        current_round (int): The current round.
    """

    def __init__(self, **kwargs: dict[str, Any]):
        self.global_eval: dict = {}  # round -> evals
        self.locals_eval: dict = {}  # round -> {client_id -> evals}
        self.prefit_eval: dict = {}  # round -> {client_id -> evals}
        self.postfit_eval: dict = {}  # round -> {client_id -> evals}
        self.locals_eval_summary: dict = {}  # round -> evals (mean across clients)
        self.prefit_eval_summary: dict = {}  # round -> evals (mean across clients)
        self.postfit_eval_summary: dict = {}  # round -> evals (mean across clients)
        self.comm_costs: dict = {0: 0}
        self.mem_costs: dict = {}
        self.current_round: int = 0

    def log(self, message: str) -> None:
        """Log a message.

        Args:
            message (str): The message to log.
        """
        rich.print(message)

    def add_scalar(self, key: Any, value: float, round: int) -> None:
        """Add a scalar to the logger.

        Args:
            key (Any): The key of the scalar.
            value (float): The value of the scalar.
            round (int): The round.
        """
        pass

    def add_scalars(self, key: Any, values: dict[str, float], round: int) -> None:
        """Add scalars to the logger.

        Args:
            key (Any): The main key of the scalars.
            values (dict[str, float]): The key-value pairs of the scalars.
            round (int): The round.
        """
        pass

    def pretty_log(self, data: Any, title: str) -> None:
        """Log a pretty-printed data.

        Args:
            data (Any): The data to log.
            title (str): The title of the data.
        """
        rich.print(Panel(Pretty(data, expand_all=True), title=title))

    def init(self, **kwargs: dict[str, Any]) -> None:
        """Initialize the logger.
        The initialization is done by printing the configuration in the console.

        Args:
            **kwargs: The configuration.
        """
        if kwargs:
            rich.print(Panel(Pretty(kwargs, expand_all=True), title="Configuration"))

    def start_round(self, round: int, global_model: Module) -> None:
        self.comm_costs[round] = 0
        self.current_round = round

        if round == 1 and self.comm_costs[0] > 0:
            rich.print(Panel(Pretty({"comm_costs": self.comm_costs[0]}), title=f"Round: {round-1}"))

    def end_round(self, round: int) -> None:
        stats = {}
        # Pre-fit summary
        if self.prefit_eval and round in self.prefit_eval and self.prefit_eval[round]:
            client_mean = pd.DataFrame(self.prefit_eval[round].values()).mean(
                numeric_only=True).to_dict()
            client_mean = {k: float(np.round(float(v), 5)) for k, v in client_mean.items()}
            self.prefit_eval_summary[round] = client_mean
            stats['pre-fit'] = client_mean

        # Post-fit summary
        if self.postfit_eval and round in self.postfit_eval and self.postfit_eval[round]:
            client_mean = pd.DataFrame(self.postfit_eval[round].values()).mean(
                numeric_only=True).to_dict()
            client_mean = {k: float(np.round(float(v), 5)) for k, v in client_mean.items()}
            self.postfit_eval_summary[round] = client_mean
            stats['post-fit'] = client_mean

        # Locals summary
        if self.locals_eval and round in self.locals_eval and self.locals_eval[round]:
            client_mean = pd.DataFrame(list(self.locals_eval[round].values())).mean(
                numeric_only=True).to_dict()
            client_mean = {k: float(np.round(float(v), 5)) for k, v in client_mean.items()}
            self.locals_eval_summary[round] = client_mean
            stats['locals'] = self.locals_eval_summary[round]

        # Global summary
        if self.global_eval and round in self.global_eval and self.global_eval[round]:
            stats['global'] = self.global_eval[round]

        stats['comm_cost'] = self.comm_costs[round]
        proc = psutil.Process(os.getpid())
        self.mem_costs[round] = proc.memory_full_info().uss

        if stats:
            rich.print(Panel(Pretty(stats, expand_all=True), title=f"Round: {round}"))
            rich.print(f"  Memory usage: {bytes2human(self.mem_costs[round])}" +
                       f"[{proc.memory_percent():.2f} %]")

    def client_evaluation(self,
                          round: int,
                          client_id: int,
                          phase: Literal['pre-fit', 'post-fit'],
                          evals: dict[str, float],
                          **kwargs: dict[str, Any]) -> None:

        if round == -1:
            round = self.current_round + 1
        dict_ref = self.prefit_eval if phase == 'pre-fit' else self.postfit_eval
        dict_ref[round] = {client_id: evals}
        self.postfit_eval_summary[round] = pd.DataFrame(list(dict_ref[round].values())).mean(
            numeric_only=True).to_dict()

    def server_evaluation(self,
                          round: int,
                          type: Literal['global', 'locals'],
                          evals: Union[dict[str, float], dict[int, dict[str, float]]],
                          **kwargs: dict[str, Any]) -> None:

        if type == 'global' and evals:
            self.global_eval[round] = evals
        elif type == "locals" and evals:
            self.locals_eval[round] = evals

    def message_received(self, message: Message) -> None:
        """Update the communication costs.

        Args:
            message (Message): The message received.
        """
        self.comm_costs[self.current_round] += message.size

    def finished(self, round: int) -> None:
        stats = {}

        # Pre-fit summary
        if self.prefit_eval and round in self.prefit_eval and self.prefit_eval[round]:
            client_mean = pd.DataFrame(self.prefit_eval[round].values()).mean(
                numeric_only=True).to_dict()
            client_mean = {k: float(np.round(float(v), 5)) for k, v in client_mean.items()}
            self.prefit_eval_summary[round] = client_mean
            stats['pre-fit'] = client_mean

        # Locals summary
        if self.locals_eval:
            stats['locals'] = self.locals_eval_summary[max(self.locals_eval.keys())]

        # Post-fit summary
        if self.postfit_eval:
            stats['post-fit'] = self.postfit_eval_summary[max(self.postfit_eval.keys())]

        # Global summary
        if self.global_eval:
            stats['global'] = self.global_eval[max(self.global_eval.keys())]

        if stats:
            rich.print(Panel(Pretty(stats, expand_all=True), title="Overall Performance"))

        rich.print(Panel(Pretty({"comm_costs": sum(self.comm_costs.values())}, expand_all=True),
                         title="Total communication cost"))

    def interrupted(self):
        rich.print("\n[bold italic yellow]The experiment has been interrupted by the user.")

    def save(self, path: str) -> None:
        """Save the logger's history to a JSON file.

        Args:
            path (str): The path to the JSON file.
        """
        json_to_save = {
            "perf_global": self.global_eval,
            "comm_costs": self.comm_costs,
            "perf_locals": self.locals_eval_summary,
            "perf_prefit": self.prefit_eval_summary,
            "perf_postfit": self.postfit_eval_summary
        }
        with open(path, 'w') as f:
            json.dump(json_to_save, f, indent=4)

    def close(self) -> None:
        """Close the logger."""
        pass


class TensorboardLog(Log):
    """TensorBoard logger.
    This class is used to log the performance of the global model and the communication costs during
    the federated learning process on TensorBoard

    See Also:
        For more information on TensorBoard, see the `official documentation
        <https://www.tensorflow.org/tensorboard>`_.

    Args:
        **config: The configuration for TensorBoard.
    """

    def __init__(self, **config):
        super().__init__(**config)
        ts_config = DDict(**config).exclude("name")
        if "log_dir" not in ts_config:
            exp_name = config['name']
            if exp_name.startswith("fluke.algorithms."):
                exp_name = ".".join(str(exp_name).split(".")[3:])
            ts_config.log_dir = f"./runs/{exp_name}" + "_" + time.strftime("%Y%m%dh%H%M%S")
        self._writer = SummaryWriter(**ts_config)

    def add_scalar(self, key: Any, value: float, round: int) -> None:
        return self._writer.add_scalars(key, value, round)

    def add_scalars(self, key: Any, values: dict[str, float], round: int) -> None:
        return self._writer.add_scalars(key, values, round)

    def start_round(self, round: int, global_model: Module) -> None:
        super().start_round(round, global_model)
        if round == 1 and self.comm_costs[0] > 0:
            self._writer.add_scalar("comm_costs", self.comm_costs[0], round)
        self._writer.flush()

    def _report(self, prefix: str, evals: dict[str, float], round: int) -> None:
        for metric, value in evals.items():
            self._writer.add_scalar(f"{prefix}/{metric}", value, round)
            self._writer.flush()

    def end_round(self, round: int) -> None:
        super().end_round(round)
        if round in self.global_eval:
            self._report("global", self.global_eval[round], round)
        self._writer.add_scalar("comm_costs", self.comm_costs[round], round)
        self._writer.flush()

        if self.prefit_eval_summary and round in self.prefit_eval_summary:
            self._report("pre-fit", self.prefit_eval_summary[round], round)

        if self.postfit_eval_summary and round in self.postfit_eval_summary:
            self._report("post-fit", self.postfit_eval_summary[round], round)

        if self.locals_eval_summary and round in self.locals_eval_summary:
            self._report("locals", self.locals_eval_summary[round], round)

        self._writer.flush()

    def finished(self, round: int) -> None:
        super().finished(round)
        if self.prefit_eval_summary and round in self.prefit_eval_summary:
            self._report("pre-fit", self.prefit_eval_summary[round], round)

        if self.locals_eval_summary and round in self.locals_eval_summary:
            self._report("locals", self.locals_eval_summary[round], round)

    def close(self) -> None:
        self._writer.flush()
        self._writer.close()


class WandBLog(Log):
    """Weights and Biases logger.
    This class is used to log the performance of the global model and the communication costs during
    the federated learning process on Weights and Biases.

    See Also:
        For more information on Weights and Biases, see the `Weights and Biases documentation
        <https://docs.wandb.ai/>`_.

    Args:
        **config: The configuration for Weights and Biases.
    """

    def __init__(self, **config):
        super().__init__(**config)
        self.config = config

    def init(self, **kwargs: dict[str, Any]) -> None:
        super().init(**kwargs)
        self.config["config"] = kwargs
        self.run = wandb.init(**self.config)

    def add_scalar(self, key: Any, value: float, round: int) -> None:
        return self.run.log({key: value}, step=round)

    def add_scalars(self, key: Any, values: dict[str, float], round: int) -> None:
        return self.run.log({f"{key}/{k}": v for k, v in values.items()}, step=round)

    def start_round(self, round: int, global_model: Module) -> None:
        super().start_round(round, global_model)
        if round == 1 and self.comm_costs[0] > 0:
            self.run.log({"comm_costs": self.comm_costs[0]})

    def end_round(self, round: int) -> None:
        super().end_round(round)
        if round in self.global_eval:
            self.run.log({"global": self.global_eval[round]}, step=round)
        self.run.log({"comm_cost": self.comm_costs[round]}, step=round)

        if self.prefit_eval_summary and round in self.prefit_eval_summary:
            self.run.log({"prefit": self.prefit_eval_summary[round]}, step=round)

        if self.postfit_eval_summary and round in self.postfit_eval_summary:
            self.run.log({"postfit": self.postfit_eval_summary[round]}, step=round)

        if self.locals_eval_summary and round in self.locals_eval_summary:
            self.run.log({"locals": self.locals_eval_summary[round]}, step=round)

    def finished(self, round: int) -> None:
        super().finished(round)

        if self.prefit_eval_summary and round in self.prefit_eval_summary:
            self.run.log({"prefit": self.prefit_eval_summary[round]}, step=round)

        if self.locals_eval_summary and round in self.locals_eval_summary:
            self.run.log({"locals": self.locals_eval_summary[round]}, step=round)

    def close(self) -> None:
        self.run.finish()

    def save(self, path: str) -> None:
        super().save(path)


class ClearMLLog(TensorboardLog):
    """ClearML logger.
    This class is used to log the performance of the global model and the communication costs during
    the federated learning process on ClearML.

    Note:
        The ClearML logger takes advantage of the TensorBoard logger, thus the logging also happens
        on TensorBoard. The logging folder is "./runs/{experiment_name}_{timestamp}".

    See Also:
        For more information on ClearML, see the `official documentation
        <https://clear.ml/docs/latest/docs/>`_.

    Args:
        **config: The configuration for ClearML.
    """

    def __init__(self, **config):
        super().__init__(name=config['name'])
        self.config = DDict(**config)

    def init(self, **kwargs: dict[str, Any]) -> None:
        super().init(**kwargs)
        # imported here to avoid issues with requests
        from clearml import Task
        self.task = Task.init(task_name=self.config.name, **self.config.exclude("name"))
        self.task.connect(kwargs)


def get_logger(lname: str, **kwargs: dict[str, Any]) -> Log:
    """Get a logger from its name.
    This function is used to get a logger from its name. It is used to dynamically import loggers.
    The supported loggers are the ones defined in the ``fluke.utils.log`` module.

    Args:
        lname (str): The name of the logger.
        **kwargs: The keyword arguments to pass to the logger's constructor.

    Returns:
        Log | WandBLog | ClearMLLog | TensorboardLog: The logger.
    """
    return get_class_from_str("fluke.utils.log", lname)(**kwargs)



"""Implementation of the FedAMP [FedAMP21]_ algorithm.

References:
    .. [FedAMP21] Yutao Huang, Lingyang Chu, Zirui Zhou, Lanjun Wang, Jiangchuan Liu, Jian Pei, Yong
       Zhang. Personalized Cross-Silo Federated Learning on Non-IID Data. In AAAI (2021).
       URL: https://arxiv.org/abs/2007.03797
"""
import sys
from copy import deepcopy
from typing import Any, Iterable

import torch
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from ..client import PFLClient  # NOQA
from ..comm import Message  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache  # NOQA
from ..utils.model import safe_load_state_dict  # NOQA
from . import PersonalizedFL  # NOQA


class FedAMPClient(PFLClient):
    def __init__(self,
                 index: int,
                 model: Module,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 lam: float = 0.2,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, model=model, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.hyper_params.update(lam=lam)

    def _alpha(self):
        return self.optimizer.param_groups[0]["lr"]

    def _proximal_loss(self, local_model, u_model):
        proximal_term = 0.0
        for w, w_t in zip(local_model.parameters(), u_model.parameters()):
            proximal_term += torch.norm(w - w_t)**2
        return (self.hyper_params.lam / (2 * self._alpha())) * proximal_term

    def receive_model(self) -> None:
        try:
            msg = self.channel.receive(self, self.server, msg_type="model")
            safe_load_state_dict(self.personalized_model, msg.payload.state_dict())
        except ValueError:
            pass

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)
        self.model.to(self.device)
        self.personalized_model.to(self.device)
        self.model.train()

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(
                    y_hat, y) + self._proximal_loss(self.model, self.personalized_model)
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()
            self.scheduler.step()

        running_loss /= (epochs * len(self.train_set))
        self.model.cpu()
        self.personalized_model.cpu()
        clear_cuda_cache()

        return running_loss

    def _load_from_cache(self):
        super()._load_from_cache()
        if self.model is None:
            self.model = deepcopy(self.personalized_model)


class FedAMPServer(Server):

    def __init__(self,
                 model: Module,
                 test_set: FastDataLoader,  # not used
                 clients: Iterable[PFLClient],
                 sigma: float = 0.1,
                 alpha: float = 0.1):
        super().__init__(model=model, test_set=None, clients=clients, weighted=False)
        self.hyper_params.update(
            sigma=sigma,
            alpha=alpha
        )

    def __e(self, x: float):
        return torch.exp(-x / self.hyper_params.sigma) / self.hyper_params.sigma

    def _empty_model(self):
        empty_model = deepcopy(self.model)
        for param in empty_model.parameters():
            param.data.zero_()
        return empty_model

    @torch.no_grad()
    def aggregate(self, eligible: Iterable[PFLClient], client_models: Iterable[Module]) -> None:

        client_models = list(client_models)
        for i, (client, ci_model) in enumerate(zip(eligible, client_models)):
            ui_model = self._empty_model()

            coef = torch.zeros(len(eligible))
            for j, cj_model in enumerate(client_models):
                if i != j:
                    weights_i = torch.cat([p.data.view(-1)
                                           for p in ci_model.parameters()], dim=0)
                    weights_j = torch.cat([p.data.view(-1)
                                           for p in cj_model.parameters()], dim=0)
                    sub = (weights_i - weights_j).view(-1)
                    sub = torch.dot(sub, sub)
                    coef[j] = self.hyper_params.alpha * self.__e(sub)
            coef[i] = 1 - torch.sum(coef)

            for j, cj_model in enumerate(client_models):
                for param_i, param_j in zip(ui_model.parameters(), cj_model.parameters()):
                    param_i.data += coef[j] * param_j

            self.channel.send(Message(ui_model, "model", self, inmemory=True), client)

    def broadcast_model(self, eligible: Iterable[PFLClient]) -> None:
        # Models have already been sent to clients in aggregate
        pass


class FedAMP(PersonalizedFL):

    def get_client_class(self) -> PFLClient:
        return FedAMPClient

    def get_server_class(self) -> Server:
        return FedAMPServer



"""Implementation of the [FedSGD17]_ algorithm.

References:
    .. [FedSGD17] H. Brendan McMahan, Eider Moore, Daniel Ramage, Seth Hampson, and Blaise Aguera
       y Arcas. Communication-efficient learning of deep networks from decentralized data.
       In AISTATS (2017). URL: https://arxiv.org/abs/1602.05629
"""
import sys
from typing import Any

import torch

sys.path.append(".")
sys.path.append("..")

from ..algorithms import CentralizedFL  # NOQA
from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..utils import OptimizerConfigurator  # NOQA


__all__ = [
    "FedSGDClient",
    "FedSGD"
]


class FedSGDClient(Client):
    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int = 3,
                 fine_tuning_epochs: int = 0,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, **kwargs)
        self.train_set.single_batch = True
        self.train_set.shuffle = True
        self.hyper_params.local_epochs = 1


class FedSGD(CentralizedFL):

    def get_client_class(self) -> Client:
        return FedSGDClient



"""Implementation of the FedALA [FedALA23]_ algorithm.

References:
    .. [FedALA23] Jianqing Zhang, Yang Hua, Hao Wang, Tao Song, Zhengui Xue, Ruhui Ma, Haibing Guan.
       FedALA: Adaptive Local Aggregation for Personalized Federated Learning
       In AAAI (2023). URL: https://arxiv.org/pdf/2212.01197v4

"""
import sys
from copy import deepcopy
from typing import Any

import numpy as np
import torch
from torch import nn

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..nets import EncoderHeadNet  # NOQA
from ..utils import OptimizerConfigurator  # NOQA
from ..utils.model import safe_load_state_dict  # NOQA
from . import CentralizedFL  # NOQA

__all__ = [
    "FedALAClient",
    "FedALA"
]


class FedALAClient(Client):

    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: nn.Module,
                 local_epochs: int = 3,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 ala_sample_size: float = 0.8,
                 eta: float = 1.0,
                 conergence_threshold: float = 0.001,
                 loss_window_size: int = 10,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.hyper_params.update(ala_sample_size=ala_sample_size,
                                 eta=eta,
                                 conergence_threshold=conergence_threshold,
                                 loss_window_size=loss_window_size)
        self.weights = None
        self.start_phase = True

    def adaptive_local_aggregation(self, server_model: EncoderHeadNet):

        # keep the server encoder weights
        safe_load_state_dict(self.model.encoder, server_model.encoder.state_dict())

        random_sample_loader = FastDataLoader(*self.train_set.tensors,
                                              num_labels=self.train_set.num_labels,
                                              batch_size=self.train_set.batch_size,
                                              shuffle=True,
                                              percentage=self.hyper_params.ala_sample_size)

        temp_model = deepcopy(self.model)

        self.model.to(self.device)
        temp_model.to(self.device)
        server_model.to(self.device)

        # frozen the encoder weights
        for param in temp_model.encoder.parameters():
            param.requires_grad = False

        optimizer = torch.optim.SGD(temp_model.head.parameters(), lr=0)

        if self.weights is None:
            self.weights = [torch.ones_like(p.data).to(self.device)
                            for p in self.model.head.parameters()]

        # initialize the temp model
        for param_t, param, param_g, weight in zip(temp_model.head.parameters(),
                                                   self.model.head.parameters(),
                                                   server_model.head.parameters(),
                                                   self.weights):
            param_t.data = param + (param_g - param) * weight

        converged = False
        losses = []
        while not converged and len(losses) < 100:

            for X, y in random_sample_loader:
                X, y = X.to(self.device), y.to(self.device)
                optimizer.zero_grad()
                y_hat = temp_model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()
                # optimizer.step()

                losses.append(loss.item())

                # update weight in this batch
                for param_t, param, param_g, weight in zip(temp_model.head.parameters(),
                                                           self.model.head.parameters(),
                                                           server_model.head.parameters(),
                                                           self.weights):
                    weight.data = torch.clamp(
                        weight - self.hyper_params.eta * (param_t.grad * (param_g - param)), 0, 1)

                # update temp local model in this batch
                for param_t, param, param_g, weight in zip(temp_model.head.parameters(),
                                                           self.model.head.parameters(),
                                                           server_model.head.parameters(),
                                                           self.weights):
                    param_t.data = param + (param_g - param) * weight

            # only train one epoch in the subsequent iterations
            if not self.start_phase:
                break

            # train the weight until convergence
            if len(losses) > self.hyper_params.loss_window_size:
                loss_std = np.std(losses[-self.hyper_params.loss_window_size:])
                if loss_std < self.hyper_params.conergence_threshold:
                    converged = True

        self.start_phase = False
        self.model.cpu()
        temp_model.cpu()
        server_model.cpu()
        safe_load_state_dict(self.model.head, temp_model.head.state_dict())

    def receive_model(self) -> None:
        server_model = self.channel.receive(self, self.server, msg_type="model").payload
        if self.model is None:
            self.model = server_model
        else:
            self.adaptive_local_aggregation(server_model)


class FedALA(CentralizedFL):

    def get_client_class(self) -> Client:
        return FedALAClient



"""Implementation of the FedROD [FedROD22]_ algorithm.

References:
    .. [FedROD22] Hong-You Chen and Wei-Lun Chao. On Bridging Generic and Personalized Federated
       Learning for Image Classification. In ICLR (2022).
       URL: https://openreview.net/pdf?id=I1hQbx10Kxn
"""
import sys
from copy import deepcopy
from typing import Any, Literal

import numpy as np
import torch
from torch.nn import functional as F
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler as LRScheduler

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..evaluation import Evaluator  # NOQA
from ..nets import EncoderHeadNet  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache  # NOQA
from ..utils.model import ModOpt  # NOQA
from . import CentralizedFL  # NOQA

__all__ = [
    "RODModel",
    "BalancedSoftmaxLoss",
    "FedRODClient",
    "FedROD"
]


class RODModel(torch.nn.Module):
    """Model that combines a global model and a local head.
    During the forward pass, the global model, formed by an encoder and a head, is used to extract
    the representation of the input (using the encoder). The representation is then passed to
    the local head and the global head. The output of the local head is added to the output of the
    global head and returned as the final output.

    Args:
        global_model (EncoderHeadNet): Global model.
        local_head (EncoderHeadNet): Local head.
    """

    def __init__(self, global_model: EncoderHeadNet, local_head: EncoderHeadNet):
        super().__init__()
        self.local_head = local_head
        self.global_model = global_model

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        rep = self.global_model.encoder(x)
        out_g = self.global_model.head(rep)
        out_p = self.local_head(rep.detach())
        output = out_g.detach() + out_p
        return output


class BalancedSoftmaxLoss(torch.nn.Module):
    """Compute the Balanced Softmax Loss.

    Args:
        sample_per_class (torch.Tensor): Number of samples per class.
    """

    def __init__(self, sample_per_class: torch.Tensor, reduction: Literal["mean", "sum"] = "mean"):
        super().__init__()
        self.sample_per_class = sample_per_class
        self.reduction = reduction

    def forward(self,
                y: torch.LongTensor,
                logits: torch.FloatTensor) -> torch.Tensor:
        spc = self.sample_per_class.type_as(logits)
        spc = spc.unsqueeze(0).expand(logits.shape[0], -1)
        logits = logits + spc.log()
        loss = F.cross_entropy(input=logits, target=y, reduction=self.reduction)
        return loss


class FedRODClient(Client):

    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)

        self.sample_per_class: torch.Tensor = torch.zeros(self.train_set.num_labels)
        uniq_val, uniq_count = np.unique(self.train_set.tensors[1], return_counts=True)
        for i, c in enumerate(uniq_val.tolist()):
            self.sample_per_class[c] = uniq_count[i]
        self._inner_modopt: ModOpt = ModOpt()

    @property
    def inner_model(self) -> ModOpt:
        return self._inner_modopt.model

    @inner_model.setter
    def inner_model(self, model: torch.nn.Module) -> None:
        self._inner_modopt.model = model

    @property
    def optimizer_head(self) -> Optimizer:
        return self._inner_modopt.optimizer

    @optimizer_head.setter
    def optimizer_head(self, optimizer: Optimizer) -> None:
        self._inner_modopt.optimizer = optimizer

    @property
    def scheduler_head(self) -> LRScheduler:
        return self._inner_modopt.scheduler

    @scheduler_head.setter
    def scheduler_head(self, scheduler: LRScheduler) -> None:
        self._inner_modopt.scheduler = scheduler

    def receive_model(self) -> None:
        super().receive_model()
        if self.inner_model is None:
            self.inner_model = deepcopy(self.model.head)

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs
                       else self.hyper_params.local_epochs)
        self.model.train()
        self.inner_model.train()
        self.model.to(self.device)
        self.inner_model.to(self.device)

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)
            self.optimizer_head, self.scheduler_head = self._optimizer_cfg(self.inner_model)

        bsm_loss = BalancedSoftmaxLoss(self.sample_per_class)
        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)

                rep = self.model.encoder(X)
                out_g = self.model.head(rep)
                loss = bsm_loss(y, out_g)
                self.optimizer.zero_grad()
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()

                out_p = self.inner_model(rep.detach())
                loss = self.hyper_params.loss_fn(out_g.detach() + out_p, y)
                self.optimizer_head.zero_grad()
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer_head.step()

            self.scheduler.step()
            self.scheduler_head.step()

        running_loss /= (epochs * len(self.train_set))
        self.model.cpu()
        self.inner_model.cpu()
        clear_cuda_cache()
        return running_loss

    def evaluate(self, evaluator: Evaluator, test_set: FastDataLoader) -> dict[str, float]:
        if test_set is not None and self.model is not None and self.inner_model is not None:
            return evaluator.evaluate(self._last_round,
                                      RODModel(self.model, self.inner_model),
                                      test_set,
                                      device=self.device)
        return {}


class FedROD(CentralizedFL):

    def get_client_class(self) -> Client:
        return FedRODClient



"""Implementation of the CCVR [CCVR21]_ algorithm.

References:
    .. [CCVR21] Mi Luo, Fei Chen, Dapeng Hu, Yifan Zhang, Jian Liang, Jiashi Feng. No Fear of
       Heterogeneity: Classifier Calibration for Federated Learning with Non-IID Data. In NeurIPS
       (2021). URL: https://arxiv.org/abs/2106.05001
"""
import sys
from typing import Any, Iterable

import numpy as np
import torch
from torch.nn.modules import Module

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from . import CentralizedFL  # NOQA

__all__ = [
    "CCVRClient",
    "CCVRServer",
    "CCVR"
]


class CCVRClient(Client):
    @torch.no_grad()
    def compute_mean_cov(self) -> None:
        """Computes the label-wise mean and covariance of the data. After the computation, the
        client send (through the channel) a message to the server containing the computed values and
        the number of examples per class.
        """
        self._load_from_cache()
        self.receive_model()
        self.model.to(self.device)
        list_z, list_y = [], []
        for _, (X, y) in enumerate(self.train_set):
            X, y = X.to(self.device), y.to(self.device)
            Z = self.model.forward_encoder(X)
            list_z.append(Z)
            list_y.append(y)

        Z = torch.cat(list_z, dim=0)
        Y = torch.cat(list_y, dim=0)
        n_feats = Z.shape[-1]

        classes_mean = []
        classes_cov = []
        ex_x_class = []
        for c in range(self.train_set.num_labels):
            idx = torch.where(Y == c)[0]
            if idx.shape[0] == 0:
                classes_mean.append(torch.zeros(n_feats))
                classes_cov.append(torch.zeros(n_feats, n_feats))
                ex_x_class.append(0)
                continue

            Z_c = Z[idx]
            mean_c = Z_c.mean(dim=0)
            cov_c = Z_c.t().cov(correction=0)
            classes_mean.append(mean_c.to("cpu"))
            classes_cov.append(cov_c.to("cpu"))
            ex_x_class.append(Z_c.size(0))

        payload = (classes_mean, classes_cov, ex_x_class)
        self.model.cpu()
        self.channel.send(Message(payload, "mean_cov", self, inmemory=True), self.server)
        self._save_to_cache()


class CCVRServer(Server):

    def __init__(self,
                 model: Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 weighted: bool = False,
                 lr: float = 0.1,
                 batch_size: int = 64,
                 sample_per_class: int = 100,
                 **kwargs: dict[str, Any]):
        super().__init__(model=model, test_set=test_set, clients=clients, weighted=weighted)
        self.hyper_params.update(
            lr=lr,
            batch_size=batch_size,
            sample_per_class=sample_per_class
        )

    def _compute_mean_cov(self) -> tuple[list[torch.Tensor], list[torch.Tensor]]:
        means, covs, ns = [], [], []
        for client in self.clients:
            client.compute_mean_cov()
            mean, cov, n = self.channel.receive(self, client, msg_type="mean_cov").payload
            means.append(mean)
            covs.append(cov)
            ns.append(n)

        num_classes = len(means[0])
        classes_mean = [None for _ in range(num_classes)]
        ex_x_class = [sum(n) for n in zip(*ns)]

        # loop over classes
        for c, (mu, n) in enumerate(zip(zip(*means), zip(*ns))):
            if ex_x_class[c] > 0:
                classes_mean[c] = torch.sum(torch.stack(
                    mu) * torch.tensor(n).reshape(-1, 1), dim=0) / ex_x_class[c]

        classes_cov = [None for _ in range(num_classes)]
        for c in range(num_classes):
            if ex_x_class[c] > 1:
                for k in range(self.n_clients):
                    if classes_cov[c] is None:
                        classes_cov[c] = torch.zeros_like(covs[k][c])

                    classes_cov[c] += ((ns[k][c] - 1) / (ex_x_class[c] - 1)
                                       ) * covs[k][c] + (ns[k][c] / (ex_x_class[c] - 1)) * (
                        torch.outer(means[k][c], means[k][c])
                    )

                classes_cov[c] -= (ex_x_class[c] / (ex_x_class[c] - 1)) * (
                    torch.outer(classes_mean[c], classes_mean[c])
                )

        return classes_mean, classes_cov

    def _generate_virtual_repr(self,
                               classes_mean: Iterable[torch.Tensor],
                               classes_cov: Iterable[torch.Tensor]) -> tuple[torch.Tensor,
                                                                             torch.Tensor]:
        data, targets = [], []
        for c, (mean, cov) in enumerate(zip(classes_mean, classes_cov)):
            if mean is not None and cov is not None:
                samples = np.random.multivariate_normal(
                    mean.cpu().numpy(),
                    cov.cpu().numpy(),
                    self.hyper_params.sample_per_class,
                )
                data.append(torch.tensor(samples, dtype=torch.float))
                targets.append(
                    torch.ones(
                        self.hyper_params.sample_per_class,
                        dtype=torch.long
                    ) * c
                )

        data = torch.cat(data)
        targets = torch.cat(targets)
        return data, targets

    def _calibrate(self, Z_train: torch.FloatTensor, y_train: torch.LongTensor) -> None:
        self.model.train()
        self.model.to(self.device)

        # FIXME: loss, optimizer and scheduler are fixed for now
        optimizer = torch.optim.SGD(self.model.head.parameters(), lr=self.hyper_params.lr)
        loss_fn = torch.nn.CrossEntropyLoss()
        train_set = FastDataLoader(Z_train,
                                   y_train,
                                   num_labels=len(set(y_train.cpu().numpy())),
                                   batch_size=self.hyper_params.batch_size,
                                   shuffle=True)

        for Z, y in train_set:
            Z, y = Z.to(self.device), y.to(self.device)
            optimizer.zero_grad()
            y_hat = self.model.head(Z)
            loss = loss_fn(y_hat, y)
            loss.backward()
            optimizer.step()
        self.model.cpu()

    def finalize(self) -> None:
        """In the CCVR Server, at the end of the learning the model is calibrated according to the
        data distributions of the clients.
        """
        self.broadcast_model(self.clients)
        classes_mean, classes_cov = self._compute_mean_cov()
        Z, y = self._generate_virtual_repr(classes_mean, classes_cov)
        self._calibrate(Z, y)
        super().finalize()


class CCVR(CentralizedFL):

    def get_client_class(self) -> Client:
        return CCVRClient

    def get_server_class(self) -> Server:
        return CCVRServer



"""`fluke-get` command line interface."""
import os

import requests
import rich
import typer
import yaml

app = typer.Typer()


@app.command()
def list():
    """
    List all available configuration files.
    """

    url = "https://api.github.com/repos/makgyver/fluke/contents/configs"
    response = requests.get(url, timeout=5)
    response.raise_for_status()

    configs = [file["name"].removesuffix('.yaml')
               for file in response.json() if file["name"].endswith(".yaml")]

    rich.print("[yellow bold]Available config files:[/]")
    for config in configs:
        rich.print(config)


@app.command()
def config(name: str, outdir: str = typer.Option("config", help="Output directory")):
    """
    Get a configuration file by name.
    """

    url = f"https://raw.githubusercontent.com/makgyver/fluke/main/configs/{name}.yaml"

    rich.print(f"Getting config file from {url} ...")
    response = requests.get(url, timeout=5)

    if response.status_code != 200:
        rich.print(f"[red][Error]:[/] [yellow]config file {name} not found.[/]")
        return

    config = yaml.safe_load(response.text)

    if not config:
        rich.print("Config file is empty.")
        return

    # if outdir does not exist, create it
    if not os.path.exists(outdir):
        rich.print(f"Creating output directory {outdir} ...")
        os.makedirs(outdir)

    if os.path.exists(f"{outdir}/{name}.yaml"):
        rich.print(
            "[red][Error]:[/] [yellow]refusing to overwrite existing config" +
            f"file {outdir}/{name}.yaml. Please rename it or delete it.[/]")
        return

    rich.print(f"Saving config file to {outdir}/{name}.yaml ...")
    with open(f"{outdir}/{name}.yaml", "w", encoding="utf8") as f:
        yaml.dump(config, f)


def main():
    app()


if __name__ == "__main__":
    app()



"""Implementation of the [SuPerFed22]_ algorithm.

References:
    .. [SuPerFed22] Seok-Ju Hahn, Minwoo Jeong, and Junghye Lee. Connecting Low-Loss Subspace
       for Personalized Federated Learning. In KDD (2022). URL: https://arxiv.org/abs/2109.07628v3
"""
import sys
from typing import Any
from copy import deepcopy

import numpy as np
from torch.nn.modules import Module

sys.path.append(".")
sys.path.append("..")

from ..algorithms import PersonalizedFL  # NOQA
from ..client import PFLClient  # NOQA
from ..data import FastDataLoader  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache, retrieve_obj  # NOQA
from ..utils.model import (get_global_model_dict, get_local_model_dict, mix_networks,  # NOQA
                           set_lambda_model, safe_load_state_dict, ModOpt)

__all__ = [
    "SuPerFedClient",
    "SuPerFed"
]


class SuPerFedClient(PFLClient):

    def __init__(self,
                 index: int,
                 model: Module,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: Module,
                 local_epochs: int = 3,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 mode: str = "global",
                 start_mix: int = 10,
                 mu: float = 0.1,
                 nu: float = 0.1,
                 **kwargs: dict[str, Any]):
        assert mode in ["mm", "lm"]

        super().__init__(index=index, model=mix_networks(model, deepcopy(model), 0),
                         train_set=train_set, test_set=test_set, optimizer_cfg=optimizer_cfg,
                         loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.hyper_params.update(
            mode=mode,
            start_mix=start_mix,
            mu=mu,
            nu=nu
        )

    def receive_model(self) -> None:
        msg = self.channel.receive(self, self.server, msg_type="model")
        if self.model is None:
            self.model = msg.payload
        else:
            safe_load_state_dict(self.model, msg.payload.state_dict())

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)

        if self.server.rounds >= self.hyper_params.start_mix:
            local_dict = {k: v for k, v in self.personalized_model.state_dict().items()
                          if '_local' in k}
        else:
            local_dict = {k + "_local": v for k,
                          v in self.model.state_dict().items()}
        self.personalized_model.load_state_dict({**self.model.state_dict(), **local_dict})

        self.personalized_model.train()
        self.personalized_model.to(self.device)

        if self.pers_optimizer is None:
            self.pers_optimizer, self.pers_scheduler = self._optimizer_cfg(self.personalized_model)

        if self.hyper_params.mu > 0:
            prev_global_model = self.model.to(self.device)
            for param in prev_global_model.parameters():
                param.requires_grad = False

        running_loss = 0.0
        for _ in range(epochs):
            loss = None
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)

                if self.server.rounds >= self.hyper_params.start_mix:
                    if self.hyper_params.mode == "mm":
                        set_lambda_model(self.personalized_model, np.random.uniform(0.0, 1.0))
                    elif self.hyper_params.mode == "lm":
                        set_lambda_model(self.personalized_model, None, layerwise=True)

                self.pers_optimizer.zero_grad()
                y_hat = self.personalized_model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)

                if self.hyper_params.mu > 0:
                    prox_term = 0.0
                    for name, param in self.personalized_model.named_parameters():
                        if '_local' not in name:
                            continue
                        internal_param = self.personalized_model.get_parameter(
                            name.replace('_local', ''))
                        prev_param = prev_global_model.get_parameter(name.replace('_local', ''))
                        prox_term += (internal_param - prev_param).norm(2)
                    loss += self.hyper_params.mu * prox_term

                if self.server.rounds >= self.hyper_params.start_mix:
                    numerator, norm_1, norm_2 = 0, 0, 0
                    for name, param_l in self.personalized_model.named_parameters():
                        if '_local' not in name:
                            continue
                        param_g = self.personalized_model.get_parameter(name.replace('_local', ''))
                        numerator += (param_g * param_l).add(1e-6).sum()
                        norm_1 += param_g.pow(2).sum()
                        norm_2 += param_l.pow(2).sum()
                    cos_sim = numerator.pow(2).div(norm_1 * norm_2)
                    loss += self.hyper_params.nu * cos_sim

                loss.backward()
                self._clip_grads(self.personalized_model)
                self.pers_optimizer.step()
                running_loss += loss.item()

            self.pers_scheduler.step()

        running_loss /= (epochs * len(self.train_set))
        self.personalized_model.cpu()
        clear_cuda_cache()

        set_lambda_model(self.personalized_model, 1.0)
        self.model.load_state_dict(get_global_model_dict(self.personalized_model))

        return running_loss


class SuPerFed(PersonalizedFL):

    def get_client_class(self) -> PFLClient:
        return SuPerFedClient



import sys

import torch
import pytest

sys.path.append(".")
sys.path.append("..")

from fluke.data import FastDataLoader  # NOQA
from fluke.evaluation import ClassificationEval, Evaluator  # NOQA


def test_classification_eval():
    loader = FastDataLoader(torch.FloatTensor([[1, 2, 3],
                                               [4, 5, 6],
                                               [7, 8, 9],
                                               [10, 11, 12]]),
                            torch.LongTensor([0, 1, 0, 2]),
                            num_labels=2,
                            batch_size=1,
                            shuffle=False,
                            skip_singleton=False)

    clf_eval = ClassificationEval(eval_every=1, n_classes=3)

    assert clf_eval.n_classes == 3
    assert clf_eval.eval_every == 1

    # model that only classifies with 0
    class ModelZero(torch.nn.Module):
        def forward(self, x):
            return torch.zeros(x.shape[0], 3)

    # model that only classifies with 1
    class ModelPerfect(torch.nn.Module):
        def forward(self, x):
            if torch.allclose(x, torch.FloatTensor([1, 2, 3])):
                return torch.FloatTensor([[1, 0, 0]])
            elif torch.allclose(x, torch.FloatTensor([4, 5, 6])):
                return torch.FloatTensor([[0, 1, 0]])
            elif torch.allclose(x, torch.FloatTensor([7, 8, 9])):
                return torch.FloatTensor([[1, 0, 0]])
            elif torch.allclose(x, torch.FloatTensor([10, 11, 12])):
                return torch.FloatTensor([[0, 0, 1]])

    # test with model that only classifies with 0
    zero_eval = clf_eval.evaluate(1, ModelZero(), loader, loss_fn=torch.nn.CrossEntropyLoss())
    perfect_eval = clf_eval.evaluate(1, ModelPerfect(), loader, loss_fn=torch.nn.CrossEntropyLoss())

    assert zero_eval["accuracy"] == 0.5
    assert zero_eval["micro_precision"] == 0.5
    assert zero_eval["micro_recall"] == 0.5
    assert zero_eval["micro_f1"] == 0.5
    assert zero_eval["macro_precision"] == 0.16667
    assert zero_eval["macro_recall"] == 0.33333
    assert zero_eval["macro_f1"] == 0.22222
    assert zero_eval["loss"] > 1

    assert perfect_eval["accuracy"] == 1
    assert perfect_eval["micro_precision"] == 1
    assert perfect_eval["micro_recall"] == 1
    assert perfect_eval["micro_f1"] == 1
    assert perfect_eval["macro_precision"] == 1
    assert perfect_eval["macro_recall"] == 1
    assert perfect_eval["macro_f1"] == 1
    assert perfect_eval["loss"] < 1

    assert clf_eval.evaluate(1, None, None) == {}

    assert str(clf_eval) == "ClassificationEval(eval_every=1, n_classes=3)" + \
        "[accuracy, precision, recall, f1]"
    assert repr(clf_eval) == str(clf_eval)

    class E(Evaluator):

        def evaluate(self, round, model, eval_data_loader, loss_fn):
            return super().evaluate(round, model, eval_data_loader, loss_fn)

    evaluator = E(1)
    with pytest.raises(NotImplementedError):
        evaluator.evaluate(1, None, None, None)

    with pytest.raises(NotImplementedError):
        evaluator(1, None, None, None)

    clf_eval = ClassificationEval(eval_every=3, n_classes=3)
    ev = clf_eval.evaluate(2, ModelPerfect(), loader, loss_fn=torch.nn.CrossEntropyLoss())
    assert ev == {}


if __name__ == "__main__":
    test_classification_eval()
    # 97% coverage for fluke/evaluation.py
